# pylint: disable=line-too-long
# fmt: off

import json
from enum import Enum
from pathlib import Path
from typing import Dict, List, Union

import pytest

from starknet_py.net.models.typed_data import Revision
from starknet_py.tests.e2e.fixtures.constants import TYPED_DATA_DIR
from starknet_py.utils.typed_data import (
    BasicType,
    Domain,
    Parameter,
    PresetType,
    StandardParameter,
    TypedData,
    encode_bool,
    encode_i128,
    encode_u128,
    parse_felt,
)


class CasesRev0(Enum):
    TD = "typed_data_rev_0_example.json"
    TD_STRING = "typed_data_rev_0_long_string_example.json"
    TD_FELT_ARR = "typed_data_rev_0_felt_array_example.json"
    TD_STRUCT_ARR = "typed_data_rev_0_struct_array_example.json"
    TD_STRUCT_MERKLE_TREE = "typed_data_rev_0_struct_merkletree_example.json"


class CasesRev1(Enum):
    TD = "typed_data_rev_1_example.json"
    TD_FELT_MERKLE_TREE = "typed_data_rev_1_felt_merkletree_example.json"
    TD_BASIC_TYPES = "typed_data_rev_1_basic_types_example.json"
    TD_PRESET_TYPES = "typed_data_rev_1_preset_types_example.json"
    TD_ENUM = "typed_data_rev_1_enum_example.json"


def loaded_typed_data(file_name: str) -> TypedData:
    """
    Load TypedData object from file
    """
    file_path = TYPED_DATA_DIR / file_name

    text = Path(file_path).read_text("utf-8")
    typed_data = json.loads(text)

    return TypedData.from_dict(typed_data)


@pytest.mark.parametrize(
    "value, result",
    [(123, "0x7b"), ("123", "0x7b"), ("0x7b", "0x7b"), ("short_string", "0x73686f72745f737472696e67")],
)
def test_parse_felt(value, result):
    assert parse_felt(value) == int(result, 16)


@pytest.mark.parametrize(
    "example, type_name, encoded_type",
    [
        (CasesRev0.TD, "Mail", "Mail(from:Person,to:Person,contents:felt)Person(name:felt,wallet:felt)"),
        (CasesRev0.TD_STRUCT_MERKLE_TREE, "Session", "Session(key:felt,expires:felt,root:merkletree)"),
        (CasesRev0.TD_FELT_ARR, "Mail",
         "Mail(from:Person,to:Person,felts_len:felt,felts:felt*)Person(name:felt,wallet:felt)"),
        (CasesRev0.TD_STRING, "Mail",
         "Mail(from:Person,to:Person,contents:String)Person(name:felt,wallet:felt)String(len:felt,data:felt*)"),
        (CasesRev0.TD_STRUCT_ARR, "Mail",
         "Mail(from:Person,to:Person,posts_len:felt,posts:Post*)Person(name:felt,wallet:felt)Post(title:felt,content:felt)"),
        (CasesRev1.TD, "Mail",
         """"Mail"("from":"Person","to":"Person","contents":"felt")"Person"("name":"felt","wallet":"felt")"""),
        (CasesRev1.TD_BASIC_TYPES, "Example",
         """"Example"("n0":"felt","n1":"bool","n2":"string","n3":"selector","n4":"u128","n5":"i128","n6":"ContractAddress","n7":"ClassHash","n8":"timestamp","n9":"shortstring")"""),
        (CasesRev1.TD_ENUM, "Example",
         """"Example"("someEnum":"MyEnum")"MyEnum"("Variant 1":(),"Variant 2":("u128","u128*"),"Variant 3":("u128"))"""),
        (CasesRev1.TD_FELT_MERKLE_TREE, "Example", """"Example"("value":"felt","root":"merkletree")""")
    ],
)
def test_encode_type(example, type_name, encoded_type):
    typed_data = loaded_typed_data(example.value)
    res = typed_data._encode_type(type_name)  # pylint: disable=protected-access
    assert res == encoded_type


# The expected hashes here and in tests below were calculated using starknet.js (https://github.com/0xs34n/starknet.js)
@pytest.mark.parametrize(
    "example, type_name, type_hash",
    [
        (CasesRev0.TD, "StarkNetDomain", "0x1bfc207425a47a5dfa1a50a4f5241203f50624ca5fdf5e18755765416b8e288"),
        (CasesRev0.TD, "Person", "0x2896dbe4b96a67110f454c01e5336edc5bbc3635537efd690f122f4809cc855"),
        (CasesRev0.TD, "Mail", "0x13d89452df9512bf750f539ba3001b945576243288137ddb6c788457d4b2f79"),
        (CasesRev0.TD_STRING, "String", "0x1933fe9de7e181d64298eecb44fc43b4cec344faa26968646761b7278df4ae2"),
        (CasesRev0.TD_STRING, "Mail", "0x1ac6f84a5d41cee97febb378ddabbe1390d4e8036df8f89dee194e613411b09"),
        (CasesRev0.TD_FELT_ARR, "Mail", "0x5b03497592c0d1fe2f3667b63099761714a895c7df96ec90a85d17bfc7a7a0"),
        (CasesRev0.TD_STRUCT_ARR, "Post", "0x1d71e69bf476486b43cdcfaf5a85c00bb2d954c042b281040e513080388356d"),
        (CasesRev0.TD_STRUCT_ARR, "Mail", "0x873b878e35e258fc99e3085d5aaad3a81a0c821f189c08b30def2cde55ff27"),
        (CasesRev0.TD_STRUCT_MERKLE_TREE, "Session",
         "0x1aa0e1c56b45cf06a54534fa1707c54e520b842feb21d03b7deddb6f1e340c"),
        (CasesRev0.TD_STRUCT_MERKLE_TREE, "Policy",
         "0x2f0026e78543f036f33e26a8f5891b88c58dc1e20cbbfaf0bb53274da6fa568"),
        (CasesRev1.TD, "StarknetDomain", "0x1ff2f602e42168014d405a94f75e8a93d640751d71d16311266e140d8b0a210"),
        (CasesRev1.TD, "Person", "0x30f7aa21b8d67cb04c30f962dd29b95ab320cb929c07d1605f5ace304dadf34"),
        (CasesRev1.TD, "Mail", "0x560430bf7a02939edd1a5c104e7b7a55bbab9f35928b1cf5c7c97de3a907bd"),
        (
                CasesRev1.TD_BASIC_TYPES, "Example",
                "0x1f94cd0be8b4097a41486170fdf09a4cd23aefbc74bb2344718562994c2c111"),
        (CasesRev1.TD_PRESET_TYPES, "Example", "0x1a25a8bb84b761090b1fadaebe762c4b679b0d8883d2bedda695ea340839a55"),
        (CasesRev1.TD_ENUM, "Example", "0x380a54d417fb58913b904675d94a8a62e2abc3467f4b5439de0fd65fafdd1a8"),
        (CasesRev1.TD_FELT_MERKLE_TREE, "Example",
         "0x160b9c0e8a7c561f9c5d9e3cc2990a1b4d26e94aa319e9eb53e163cd06c71be"),
    ],
)
def test_type_hash(example, type_name, type_hash):
    typed_data = loaded_typed_data(example.value)
    res = typed_data.type_hash(type_name)
    assert hex(res) == type_hash


@pytest.mark.parametrize(
    "example, type_name, attr_name, struct_hash",
    [
        (CasesRev0.TD, "StarkNetDomain", "domain",
         "0x54833b121883a3e3aebff48ec08a962f5742e5f7b973469c1f8f4f55d470b07"),
        (CasesRev0.TD, "Mail", "message", "0x4758f1ed5e7503120c228cbcaba626f61514559e9ef5ed653b0b885e0f38aec"),
        (CasesRev0.TD_STRING, "Mail", "message",
         "0x1d16b9b96f7cb7a55950b26cc8e01daa465f78938c47a09d5a066ca58f9936f"),
        (CasesRev0.TD_FELT_ARR, "Mail", "message",
         "0x26186b02dddb59bf12114f771971b818f48fad83c373534abebaaa39b63a7ce"),
        (CasesRev0.TD_STRUCT_ARR, "Mail", "message",
         "0x5650ec45a42c4776a182159b9d33118a46860a6e6639bb8166ff71f3c41eaef"),
        (CasesRev0.TD_STRUCT_MERKLE_TREE, "Session", "message",
         "0x73602062421caf6ad2e942253debfad4584bff58930981364dcd378021defe8"),
        (CasesRev1.TD, "StarknetDomain", "domain",
         "0x555f72e550b308e50c1a4f8611483a174026c982a9893a05c185eeb85399657"),
        (CasesRev1.TD_PRESET_TYPES, "Example", "message",
         "0x74fba3f77f8a6111a9315bac313bf75ecfa46d1234e0fda60312fb6a6517667"),
        (CasesRev1.TD_ENUM, "Example", "message",
         "0x3d4384ff5cec32b86462e89f5a803b55ff0048c4f5a10ba9d6cd381317d9c3"),
        (CasesRev1.TD_FELT_MERKLE_TREE, "Example", "message",
         "0x40ef40c56c0469799a916f0b7e3bc4f1bbf28bf659c53fb8c5ee4d8d1b4f5f0")
    ],
)
def test_struct_hash(example, type_name, attr_name, struct_hash):
    typed_data = loaded_typed_data(example.value)
    data = getattr(typed_data, attr_name)
    if isinstance(data, Domain):
        data = data.to_dict()
    res = typed_data.struct_hash(type_name, data)
    assert hex(res) == struct_hash


@pytest.mark.parametrize(
    "example, account_address, msg_hash",
    [
        (CasesRev0.TD, "0xcd2a3d9f938e13cd947ec05abc7fe734df8dd826",
         "0x6fcff244f63e38b9d88b9e3378d44757710d1b244282b435cb472053c8d78d0"),
        (CasesRev0.TD_STRING, "0xcd2a3d9f938e13cd947ec05abc7fe734df8dd826",
         "0x691b977ee0ee645647336f01d724274731f544ad0d626b078033d2541ee641d"),
        (CasesRev0.TD_FELT_ARR, "0xcd2a3d9f938e13cd947ec05abc7fe734df8dd826",
         "0x30ab43ef724b08c3b0a9bbe425e47c6173470be75d1d4c55fd5bf9309896bce"),
        (CasesRev0.TD_STRUCT_ARR, "0xcd2a3d9f938e13cd947ec05abc7fe734df8dd826",
         "0x5914ed2764eca2e6a41eb037feefd3d2e33d9af6225a9e7fe31ac943ff712c"),
        (CasesRev0.TD_STRUCT_MERKLE_TREE, "0xcd2a3d9f938e13cd947ec05abc7fe734df8dd826",
         "0x5d28fa1b31f92e63022f7d85271606e52bed89c046c925f16b09e644dc99794"),
        (CasesRev1.TD, "0xcd2a3d9f938e13cd947ec05abc7fe734df8dd826",
         "0x7f6e8c3d8965b5535f5cc68f837c04e3bbe568535b71aa6c621ddfb188932b8"),
        (CasesRev1.TD_BASIC_TYPES, "0xcd2a3d9f938e13cd947ec05abc7fe734df8dd826",
         "0x2d80b87b8bc32068247c779b2ef0f15f65c9c449325e44a9df480fb01eb43ec"),
        (CasesRev1.TD_PRESET_TYPES, "0xcd2a3d9f938e13cd947ec05abc7fe734df8dd826",
         "0x185b339d5c566a883561a88fb36da301051e2c0225deb325c91bb7aa2f3473a"),
        (CasesRev1.TD_ENUM, "0xcd2a3d9f938e13cd947ec05abc7fe734df8dd826",
         "0x3df10475ad5a8f49db4345a04a5b09164d2e24b09f6e1e236bc1ccd87627cc"),
        (CasesRev1.TD_FELT_MERKLE_TREE, "0xcd2a3d9f938e13cd947ec05abc7fe734df8dd826",
         "0x4f706783e0d7d0e61433d41343a248a213e9ab341d50ba978dfc055f26484c9")
    ],
)
def test_message_hash(example, account_address, msg_hash):
    typed_data = loaded_typed_data(example.value)
    res = typed_data.message_hash(int(account_address, 16))
    assert hex(res) == msg_hash


domain_type_v0 = {
    "StarkNetDomain": [
        StandardParameter(name="name", type="felt"),
        StandardParameter(name="version", type="felt"),
        StandardParameter(name="chainId", type="felt"),
    ]
}

domain_type_v1: Dict[str, List[Parameter]] = {
    "StarknetDomain": [
        StandardParameter(name="name", type="shortstring"),
        StandardParameter(name="version", type="shortstring"),
        StandardParameter(name="chainId", type="shortstring"),
        StandardParameter(name="revision", type="shortstring"),
    ]
}

domain_v0 = Domain(
    name="DomainV0",
    version="1",
    chain_id=1234,
)

domain_v1 = Domain(
    name="DomainV1",
    version="1",
    chain_id="1234",
    revision=Revision.V1,
)


def _make_typed_data(included_type: str, revision: Revision):
    domain_type, domain = (domain_type_v0, domain_v0) if revision == Revision.V0 else (
        domain_type_v1, domain_v1)

    types = {**domain_type, included_type: []}
    message = {included_type: 1}

    return TypedData(
        types=types,
        primary_type=included_type,
        domain=domain,
        message=message,
    )


@pytest.mark.parametrize(
    "included_type, revision",
    [
        ("", Revision.V1),
        ("myType*", Revision.V1),
        ("(myType)", Revision.V1),
        ("()", Revision.V1),
    ],
)
def test_invalid_type_names(included_type: str, revision: Revision):
    with pytest.raises(ValueError):
        _make_typed_data(included_type, revision)


@pytest.mark.parametrize(
    "included_type, revision",
    [
        (BasicType.FELT, Revision.V0),
        (BasicType.STRING, Revision.V0),
        (BasicType.SELECTOR, Revision.V0),
        (BasicType.MERKLE_TREE, Revision.V0),
        (BasicType.BOOL, Revision.V0),
        (BasicType.FELT, Revision.V1),
        (BasicType.STRING, Revision.V1),
        (BasicType.SELECTOR, Revision.V1),
        (BasicType.MERKLE_TREE, Revision.V1),
        (BasicType.BOOL, Revision.V1),
        (BasicType.SHORT_STRING, Revision.V1),
        (BasicType.U128, Revision.V1),
        (BasicType.I128, Revision.V1),
        (BasicType.TIMESTAMP, Revision.V1),
        (BasicType.CONTRACT_ADDRESS, Revision.V1),
        (BasicType.CLASS_HASH, Revision.V1),
        (BasicType.ENUM, Revision.V1)
    ],
)
def test_basic_types_redefinition(included_type: BasicType, revision: Revision):
    with pytest.raises(ValueError, match=fr"Types must not contain basic types. \[{included_type.value}\] was found."):
        _make_typed_data(included_type.value, revision)


@pytest.mark.parametrize(
    "included_type, revision",
    [
        (PresetType.U256, Revision.V1),
        (PresetType.TOKEN_AMOUNT, Revision.V1),
        (PresetType.NFT_ID, Revision.V1),
    ],
)
def test_preset_types_redefinition(included_type: PresetType, revision: Revision):
    with pytest.raises(ValueError, match=fr"Types must not contain preset types. \[{included_type.value}\] was found."):
        _make_typed_data(included_type.value, revision)


def test_custom_type_definition():
    _make_typed_data("myType", Revision.V0)


@pytest.mark.parametrize(
    "revision",
    list(Revision),
)
def test_missing_domain_type(revision: Revision):
    domain = domain_v0 if revision == Revision.V0 else domain_v1

    with pytest.raises(ValueError, match=f"Types must contain '{domain.separator_name}'."):
        TypedData(
            types={},
            primary_type="felt",
            domain=domain,
            message={},
        )


def test_dangling_type():
    with pytest.raises(ValueError, match=r"Dangling types are not allowed. Unreferenced type \[dangling\] was found."):
        TypedData(
            types={
                **domain_type_v1,
                "dangling": [],
                "mytype": []
            },
            primary_type="mytype",
            domain=domain_v1,
            message={"mytype": 1},
        )


def test_missing_dependency():
    typed_data = TypedData(
        types={
            **domain_type_v1,
            "house": [StandardParameter(name="fridge", type="ice cream")]
        },
        primary_type="house",
        domain=domain_v1,
        message={"fridge": 1},
    )

    with pytest.raises(ValueError, match=r"Type \[ice cream\] is not defined in types."):
        typed_data.struct_hash("house", {"fridge": 1})


@pytest.mark.parametrize(
    "value, expected",
    [
        (True, 1),
        (False, 0),
        ("true", 1),
        ("false", 0),
        ("0x1", 1),
        ("0x0", 0),
        ("1", 1),
        ("0", 0),
        (1, 1),
        (0, 0)

    ]
)
def test_encode_bool(value: Union[bool, str, int], expected: int):
    assert encode_bool(value) == expected


@pytest.mark.parametrize(
    "value",
    [
        -2,
        2,
        "-2",
        "2",
        "0x123",
        "anyvalue",
    ]
)
def test_encode_invalid_bool(value: Union[bool, str, int]):
    with pytest.raises(ValueError, match=fr"Expected boolean value, got \[{value}\]."):
        encode_bool(value)


@pytest.mark.parametrize(
    "value, expected",
    [
        (0, 0),
        (1, 1),
        (1000000, 1000000),
        ("0x0", 0),
        ("0x1", 1),
        ("0x64", 100),
        (2 ** 128 - 1, 2 ** 128 - 1),
    ]
)
def test_encode_u128(value: Union[str, int], expected: str):
    assert encode_u128(value) == expected


@pytest.mark.parametrize(
    "value",
    [
        -1,
        "-1",
        1.23,
        -1.23,
        "1.23",
        "-1.23",
        "example",
        "0xwrong",
        2 ** 128,
        hex(2 ** 128),
    ]
)
def test_encode_invalid_u128(value: Union[str, int]):
    with pytest.raises(ValueError):
        encode_u128(value)


@pytest.mark.parametrize(
    "value, expected",
    [
        (0, 0),
        (1, 1),
        (1000000, 1000000),
        ("0x0", 0),
        ("0x1", 1),
        ("0x64", 100),
        (2 ** 127 - 1, 2 ** 127 - 1),
        (-1, 3618502788666131213697322783095070105623107215331596699973092056135872020480),
        (-1000000, 3618502788666131213697322783095070105623107215331596699973092056135871020481),
        (-(2 ** 127), 3618502788666131213697322783095070105452966031871127468241404752419987914753),
    ]
)
def test_encode_i128(value: Union[str, int], expected: str):
    assert encode_i128(value) == expected


@pytest.mark.parametrize(
    "value",
    [
        "example",
        "0xwrong",
        1.23,
        -1.23,
        "1.23",
        "-1.23",
        -2 ** 127 - 1,
        2 ** 127,
        str(-2 ** 127 - 1),
        str(2 ** 127),

    ]
)
def test_encode_invalid_i128(value: Union[str, int]):
    with pytest.raises(ValueError):
        encode_i128(value)



import re

import pytest

from starknet_py.serialization.data_serializers.uint256_serializer import (
    Uint256Serializer,
)
from starknet_py.serialization.errors import InvalidTypeException, InvalidValueException

serializer = Uint256Serializer()
SHIFT = 2**128
MAX_U128 = SHIFT - 1


@pytest.mark.parametrize(
    "value, serialized_value",
    [
        (123 + 456 * SHIFT, [123, 456]),
        (
            21323213211421424142 + 347932774343 * SHIFT,
            [21323213211421424142, 347932774343],
        ),
        (0, [0, 0]),
        (MAX_U128, [MAX_U128, 0]),
        (MAX_U128 * SHIFT, [0, MAX_U128]),
        (MAX_U128 + MAX_U128 * SHIFT, [MAX_U128, MAX_U128]),
    ],
)
def test_valid_values(value, serialized_value):
    deserialized = serializer.deserialize(serialized_value)
    assert deserialized == value

    serialized = serializer.serialize(value)
    assert serialized == serialized_value

    assert serialized_value == serializer.serialize(
        {"low": serialized_value[0], "high": serialized_value[1]}
    )


def test_deserialize_invalid_values():
    # We need to escape braces
    low_error_message = re.escape(
        "Error at path 'low': expected value in range [0;2**128)"
    )
    with pytest.raises(InvalidValueException, match=low_error_message):
        serializer.deserialize([MAX_U128 + 1, 0])
    with pytest.raises(InvalidValueException, match=low_error_message):
        serializer.deserialize([MAX_U128 + 1, MAX_U128 + 1])
    with pytest.raises(InvalidValueException, match=low_error_message):
        serializer.deserialize([-1, 0])

    high_error_message = re.escape(
        "Error at path 'high': expected value in range [0;2**128)"
    )
    with pytest.raises(InvalidValueException, match=high_error_message):
        serializer.deserialize([0, MAX_U128 + 1])
    with pytest.raises(InvalidValueException, match=high_error_message):
        serializer.deserialize([0, -1])


def test_serialize_invalid_int_value():
    error_message = re.escape("Error: Uint256 is expected to be in range [0;2**256)")
    with pytest.raises(InvalidValueException, match=error_message):
        serializer.serialize(2**256)
    with pytest.raises(InvalidValueException, match=error_message):
        serializer.serialize(-1)


def test_serialize_invalid_dict_values():
    low_error_message = re.escape(
        "Error at path 'low': expected value in range [0;2**128)"
    )
    with pytest.raises(InvalidValueException, match=low_error_message):
        serializer.serialize({"low": -1, "high": 12324})
    with pytest.raises(InvalidValueException, match=low_error_message):
        serializer.serialize({"low": MAX_U128 + 1, "high": 4543535})

    high_error_message = re.escape(
        "Error at path 'high': expected value in range [0;2**128)"
    )
    with pytest.raises(InvalidValueException, match=high_error_message):
        serializer.serialize({"low": 652432, "high": -1})
    with pytest.raises(InvalidValueException, match=high_error_message):
        serializer.serialize({"low": 0, "high": MAX_U128 + 1})


def test_invalid_type():
    error_message = re.escape(
        "Error: expected int or dict, received 'wololoo' of type '<class 'str'>'."
    )
    with pytest.raises(InvalidTypeException, match=error_message):
        serializer.serialize("wololoo")  # type: ignore



import pytest

from starknet_py.net.client_models import (
    BlockStatus,
    L1DAMode,
    PreConfirmedStarknetBlock,
    PreConfirmedStarknetBlockWithTxHashes,
    StarknetBlock,
    StarknetBlockWithReceipts,
    StarknetBlockWithTxHashes,
)


@pytest.mark.asyncio
async def test_pre_confirmed_block(account):
    blk = await account.client.get_block(block_number="pre_confirmed")
    assert blk.transactions is not None
    assert isinstance(blk, PreConfirmedStarknetBlock)


@pytest.mark.asyncio
async def test_latest_block(account):
    blk = await account.client.get_block(block_number="latest")
    assert blk.block_hash
    assert blk.transactions is not None
    assert isinstance(blk, StarknetBlock)


@pytest.mark.asyncio
async def test_block_with_tx_hashes_pre_confirmed(account):
    blk = await account.client.get_block_with_tx_hashes(block_number="pre_confirmed")

    assert isinstance(blk, PreConfirmedStarknetBlockWithTxHashes)
    assert isinstance(blk.transactions, list)


@pytest.mark.asyncio
async def test_block_with_tx_hashes_latest(account):
    blk = await account.client.get_block_with_tx_hashes(block_number="latest")

    assert isinstance(blk, StarknetBlockWithTxHashes)
    assert isinstance(blk.transactions, list)
    assert blk.transactions is not None
    assert blk.block_hash is not None
    assert blk.parent_hash is not None
    assert blk.block_number is not None
    assert blk.new_root is not None
    assert blk.timestamp is not None
    assert blk.sequencer_address is not None
    assert blk.l1_gas_price.price_in_wei > 0
    assert blk.l1_gas_price.price_in_fri > 0
    assert blk.l1_data_gas_price.price_in_wei >= 0
    assert blk.l1_data_gas_price.price_in_fri >= 0
    assert blk.l1_da_mode in L1DAMode


@pytest.mark.asyncio
async def test_get_block_with_txs_pre_confirmed(account):
    blk = await account.client.get_block_with_txs(block_number="pre_confirmed")

    assert isinstance(blk, PreConfirmedStarknetBlock)
    assert isinstance(blk.transactions, list)


@pytest.mark.asyncio
async def test_get_block_with_txs_latest(account, map_class_hash):
    # pylint: disable=unused-argument

    blk = await account.client.get_block_with_txs(block_number="latest")

    assert isinstance(blk, StarknetBlock)
    assert isinstance(blk.transactions, list)
    assert blk.transactions[0].hash is not None
    assert blk.block_hash is not None
    assert blk.parent_hash is not None
    assert blk.block_number is not None
    assert blk.new_root is not None
    assert blk.timestamp is not None
    assert blk.sequencer_address is not None
    assert blk.l1_gas_price.price_in_wei > 0
    assert blk.l1_gas_price.price_in_fri > 0
    assert blk.l1_data_gas_price.price_in_wei >= 0
    assert blk.l1_data_gas_price.price_in_fri >= 0
    assert blk.l1_da_mode in L1DAMode


@pytest.mark.asyncio
async def test_block_with_receipts_latest(account):
    blk = await account.client.get_block_with_receipts(block_number="latest")

    assert isinstance(blk, StarknetBlockWithReceipts)
    assert isinstance(blk.transactions, list)
    assert blk.status == BlockStatus.ACCEPTED_ON_L2
    assert blk.block_hash is not None
    assert blk.parent_hash is not None
    assert blk.block_number is not None
    assert blk.new_root is not None
    assert blk.timestamp is not None
    assert blk.sequencer_address is not None
    assert blk.l1_gas_price.price_in_wei > 0
    assert blk.l1_gas_price.price_in_fri > 0
    assert blk.l1_data_gas_price.price_in_wei >= 0
    assert blk.l1_data_gas_price.price_in_fri >= 0
    assert blk.l2_gas_price.price_in_wei > 0
    assert blk.l2_gas_price.price_in_fri > 0
    assert blk.l1_da_mode in L1DAMode



import json

from marshmallow import EXCLUDE
from marshmallow import Schema as MarshmallowSchema
from marshmallow import SchemaOpts, ValidationError, fields, post_load

from starknet_py.abi.v0.schemas import ContractAbiEntrySchema
from starknet_py.net.client_models import (
    CasmClassEntryPoint,
    CasmClassEntryPointsByType,
    DeployedContract,
    DeprecatedCompiledContract,
    DeprecatedContractClass,
    EntryPoint,
    EntryPointsByType,
    SierraCompiledContract,
    SierraContractClass,
    SierraEntryPoint,
    SierraEntryPointsByType,
    SyncStatus,
)
from starknet_py.net.executable_models import CasmClass
from starknet_py.net.schemas.common import Felt, NumberAsHex
from starknet_py.net.schemas.rpc.executables_api import HintSchema
from starknet_py.utils.schema import ExcludeOpts, Schema


class SyncStatusSchema(Schema):
    starting_block_hash = Felt(data_key="starting_block_hash", required=True)
    starting_block_num = Felt(data_key="starting_block_num", required=True)
    current_block_hash = Felt(data_key="current_block_hash", required=True)
    current_block_num = Felt(data_key="current_block_num", required=True)
    highest_block_hash = Felt(data_key="highest_block_hash", required=True)
    highest_block_num = Felt(data_key="highest_block_num", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> SyncStatus:
        return SyncStatus(**data)


class ContractDiffSchema(Schema):
    address = Felt(data_key="address", required=True)
    contract_hash = Felt(data_key="contract_hash", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> DeployedContract:
        return DeployedContract(**data)


class SierraEntryPointSchema(Schema):
    selector = Felt(data_key="selector", required=True)
    function_idx = fields.Integer(data_key="function_idx", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> SierraEntryPoint:
        return SierraEntryPoint(**data)


class EntryPointSchema(Schema):
    offset = NumberAsHex(data_key="offset", required=True)
    selector = Felt(data_key="selector", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> EntryPoint:
        return EntryPoint(**data)


class SierraEntryPointsByTypeSchema(Schema):
    constructor = fields.List(
        fields.Nested(SierraEntryPointSchema()), data_key="CONSTRUCTOR", required=True
    )
    external = fields.List(
        fields.Nested(SierraEntryPointSchema()), data_key="EXTERNAL", required=True
    )
    l1_handler = fields.List(
        fields.Nested(SierraEntryPointSchema()), data_key="L1_HANDLER", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> SierraEntryPointsByType:
        return SierraEntryPointsByType(**data)


class EntryPointsByTypeSchema(Schema):
    constructor = fields.List(
        fields.Nested(EntryPointSchema()), data_key="CONSTRUCTOR", required=True
    )
    external = fields.List(
        fields.Nested(EntryPointSchema()), data_key="EXTERNAL", required=True
    )
    l1_handler = fields.List(
        fields.Nested(EntryPointSchema()), data_key="L1_HANDLER", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> EntryPointsByType:
        return EntryPointsByType(**data)


class SierraContractClassSchema(Schema):
    sierra_program = fields.List(Felt(), data_key="sierra_program", required=True)
    contract_class_version = fields.String(
        data_key="contract_class_version", required=True
    )
    entry_points_by_type = fields.Nested(
        SierraEntryPointsByTypeSchema(), data_key="entry_points_by_type", required=True
    )
    abi = fields.String(data_key="abi", required=False)

    @post_load
    def make_dataclass(self, data, **kwargs) -> SierraContractClass:
        return SierraContractClass(**data)


class ContractClassSchema(Schema):
    program = fields.Dict(
        keys=fields.String(),
        values=fields.Raw(load_default=None),
        data_key="program",
        required=True,
    )
    entry_points_by_type = fields.Nested(
        EntryPointsByTypeSchema(), data_key="entry_points_by_type", required=True
    )
    abi = fields.List(fields.Dict(), data_key="abi")

    @post_load
    def make_dataclass(self, data, **kwargs) -> DeprecatedContractClass:
        return DeprecatedContractClass(**data)


class DeprecatedContractClassSchema(Schema):
    program = fields.String(data_key="program", required=True)
    entry_points_by_type = fields.Nested(
        EntryPointsByTypeSchema(), data_key="entry_points_by_type", required=True
    )
    abi = fields.List(
        fields.Nested(ContractAbiEntrySchema(unknown=EXCLUDE)), data_key="abi"
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> DeprecatedContractClass:
        return DeprecatedContractClass(**data)


class DeprecatedCompiledContractSchema(ContractClassSchema):
    abi = fields.List(fields.Dict(), data_key="abi", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> DeprecatedCompiledContract:
        return DeprecatedCompiledContract(**data)


class CasmClassEntryPointSchema(Schema):
    selector = Felt(data_key="selector", required=True)
    offset = NumberAsHex(data_key="offset", required=True)
    builtins = fields.List(fields.String(), data_key="builtins")

    @post_load
    def make_dataclass(self, data, **kwargs) -> CasmClassEntryPoint:
        return CasmClassEntryPoint(**data)


class CasmClassEntryPointsByTypeSchema(Schema):
    constructor = fields.List(
        fields.Nested(CasmClassEntryPointSchema()),
        data_key="CONSTRUCTOR",
        required=True,
    )
    external = fields.List(
        fields.Nested(CasmClassEntryPointSchema()),
        data_key="EXTERNAL",
        required=True,
    )
    l1_handler = fields.List(
        fields.Nested(CasmClassEntryPointSchema()),
        data_key="L1_HANDLER",
        required=True,
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> CasmClassEntryPointsByType:
        return CasmClassEntryPointsByType(**data)


# TODO(#1564): `CasmClassSchema` should inherit from `Schema` and shouldn't overwrite `OPTION_CLASS`
#  once issue is resolved.
class CasmClassSchema(MarshmallowSchema):
    OPTIONS_CLASS = ExcludeOpts
    prime = NumberAsHex(data_key="prime", required=True)
    bytecode = fields.List(Felt(), data_key="bytecode", required=True)
    bytecode_segment_lengths = fields.List(
        fields.Integer(), data_key="bytecode_segment_lengths", load_default=None
    )
    hints = fields.List(
        fields.Tuple(
            (fields.Integer(), fields.List(fields.Nested(HintSchema()))),
        ),
        data_key="hints",
        required=True,
    )
    compiler_version = fields.String(data_key="compiler_version", required=True)
    entry_points_by_type = fields.Nested(
        CasmClassEntryPointsByTypeSchema(),
        data_key="entry_points_by_type",
        required=True,
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> CasmClass:
        return CasmClass(**data)


class AbiField(fields.Field):
    def _deserialize(self, value, attr, data, **kwargs):
        if isinstance(value, str):
            return value
        if isinstance(value, list) and all(isinstance(item, dict) for item in value):
            return json.dumps(value)
        raise ValidationError("Field should be str or list[dict].")


class SierraCompiledContractSchema(SierraContractClassSchema):
    abi = AbiField(data_key="abi", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> SierraCompiledContract:
        return SierraCompiledContract(**data)



from typing import Optional, Tuple, Union

from starknet_py.common import create_casm_class
from starknet_py.hash.casm_class_hash import compute_casm_class_hash
from starknet_py.hash.hash_method import HashMethod
from starknet_py.net.account.base_account import BaseAccount
from starknet_py.net.client import Client


def _extract_compiled_class_hash(
    compiled_contract_casm: Optional[str] = None,
    compiled_class_hash: Optional[int] = None,
    hash_method: HashMethod = HashMethod.BLAKE2S,
) -> int:
    if compiled_class_hash is None and compiled_contract_casm is None:
        raise ValueError(
            "For Cairo 1.0 contracts, either the 'compiled_class_hash' or the 'compiled_contract_casm' "
            "argument must be provided."
        )

    if compiled_class_hash is None:
        assert compiled_contract_casm is not None
        compiled_class_hash = compute_casm_class_hash(
            create_casm_class(compiled_contract_casm), hash_method=hash_method
        )

    return compiled_class_hash


def _unpack_provider(
    provider: Union[BaseAccount, Client]
) -> Tuple[Client, Optional[BaseAccount]]:
    """
    Get the client and optional account to be used by Contract.

    If provided with Client, returns this Client and None.
    If provided with BaseAccount, returns underlying Client and the account.
    """
    if isinstance(provider, Client):
        return provider, None

    if isinstance(provider, BaseAccount):
        return provider.client, provider

    raise ValueError("Argument provider is not of accepted type.")



from collections import OrderedDict

import pytest

from starknet_py.abi.v0 import Abi
from starknet_py.cairo.data_types import FeltType, NamedTupleType, StructType, TupleType
from starknet_py.serialization.data_serializers.array_serializer import ArraySerializer
from starknet_py.serialization.data_serializers.felt_serializer import FeltSerializer
from starknet_py.serialization.data_serializers.named_tuple_serializer import (
    NamedTupleSerializer,
)
from starknet_py.serialization.data_serializers.payload_serializer import (
    PayloadSerializer,
)
from starknet_py.serialization.data_serializers.struct_serializer import (
    StructSerializer,
)
from starknet_py.serialization.data_serializers.tuple_serializer import TupleSerializer
from starknet_py.serialization.data_serializers.uint256_serializer import (
    Uint256Serializer,
)
from starknet_py.serialization.data_serializers.uint_serializer import UintSerializer
from starknet_py.serialization.errors import InvalidTypeException
from starknet_py.serialization.factory import (
    serializer_for_event,
    serializer_for_function,
    serializer_for_type,
)
from starknet_py.serialization.function_serialization_adapter import (
    FunctionSerializationAdapterV0,
)
from starknet_py.tests.e2e.fixtures.abi_structures import (
    get_user_fn,
    pool_id_added_event,
    pool_id_struct,
    uint256_struct,
    user_struct,
)
from starknet_py.tests.e2e.fixtures.abi_v1_structures import abi_v1
from starknet_py.tests.e2e.fixtures.abi_v2_structures import abi_v2

abi = Abi(
    defined_structures={
        "Uint256": uint256_struct,
        "PoolId": pool_id_struct,
        "User": user_struct,
    },
    events={
        "PoolIdAdded": pool_id_added_event,
    },
    functions={
        "get_user_fn": get_user_fn,
    },
    constructor=None,
    l1_handler=None,
)

pool_id_serializer = StructSerializer(OrderedDict(value=Uint256Serializer()))
pool_id_serializer_v2 = StructSerializer(OrderedDict(value=UintSerializer(bits=256)))

user_serializer = StructSerializer(
    OrderedDict(
        id=Uint256Serializer(),
        name_len=FeltSerializer(),
        name=ArraySerializer(FeltSerializer()),
        pool_id=pool_id_serializer,
        favorite_numbers_triplet=ArraySerializer(FeltSerializer()),
    )
)

event_serializer = PayloadSerializer(OrderedDict(pool_id=pool_id_serializer_v2))


@pytest.mark.parametrize(
    "structure, serializer",
    (
        (abi.defined_structures["Uint256"], Uint256Serializer()),
        (abi.defined_structures["PoolId"], pool_id_serializer),
        (abi.defined_structures["User"], user_serializer),
        (abi_v2.events["PoolIdAdded"], event_serializer),
        (
            StructType(
                "structure",
                OrderedDict(
                    regular=TupleType([FeltType()]),
                    named=NamedTupleType(OrderedDict(value=FeltType())),
                ),
            ),
            StructSerializer(
                OrderedDict(
                    regular=TupleSerializer([FeltSerializer()]),
                    named=NamedTupleSerializer(OrderedDict(value=FeltSerializer())),
                )
            ),
        ),
    ),
)
def test_getting_type_serializer(structure, serializer):
    assert serializer_for_type(structure) == serializer


def test_getting_payload_serializer_v0():
    assert serializer_for_event(abi.events["PoolIdAdded"]) == PayloadSerializer(
        OrderedDict(pool_id=pool_id_serializer)
    )


def test_getting_payload_serializer_v1():
    assert serializer_for_event(abi_v1.events["PoolIdAdded"]) == PayloadSerializer(
        OrderedDict(pool_id=pool_id_serializer_v2)
    )


def test_getting_payload_serializer_v2():
    assert serializer_for_event(abi_v2.events["PoolIdAdded"]) == PayloadSerializer(
        OrderedDict(pool_id=pool_id_serializer_v2)
    )


def test_getting_function_serializer():
    assert serializer_for_function(
        abi.functions["get_user_fn"]
    ) == FunctionSerializationAdapterV0(
        inputs_serializer=PayloadSerializer(OrderedDict(id=Uint256Serializer())),
        outputs_deserializer=PayloadSerializer(OrderedDict(user=user_serializer)),
    )


def test_invalid_type():
    with pytest.raises(
        InvalidTypeException, match="Received unknown Cairo type 'test'."
    ):
        serializer_for_type("test")  # type: ignore



import pytest

from starknet_py.cairo.felt import (
    cairo_vm_range_check,
    decode_shortstring,
    encode_shortstring,
)
from starknet_py.constants import FIELD_PRIME


@pytest.mark.parametrize("value", [-1, FIELD_PRIME, FIELD_PRIME + 1, -FIELD_PRIME])
def test_invalid_cairo_vm_values(value):
    with pytest.raises(ValueError, match="is expected to be in range \\[0;"):
        cairo_vm_range_check(value)


@pytest.mark.parametrize("value", [0, 1, FIELD_PRIME - 1])
def cairo_vm_range_check_good_numbers(value):
    cairo_vm_range_check(value)


def test_encode_decode_shortstring():
    shortstring = "hello"
    assert decode_shortstring(encode_shortstring(shortstring)) == shortstring



from __future__ import annotations

from collections import OrderedDict

from starknet_py.abi.v0 import Abi
from starknet_py.cairo.data_types import ArrayType, FeltType, StructType

uint256_dict = {
    "type": "struct",
    "name": "Uint256",
    "size": 2,
    "members": [
        # low and high were switched on purpose. Parser has to use offset to order them properly.
        {"name": "high", "offset": 1, "type": "felt"},
        {"name": "low", "offset": 0, "type": "felt"},
    ],
}
uint256_struct = StructType("Uint256", OrderedDict(low=FeltType(), high=FeltType()))

pool_id_dict = {
    "type": "struct",
    "name": "PoolId",
    "size": 1,
    "members": [
        {"name": "value", "offset": 0, "type": "Uint256"},
    ],
}
pool_id_struct = StructType("PoolId", OrderedDict(value=uint256_struct))

user_dict = {
    "type": "struct",
    "name": "User",
    "size": 4,
    "members": [
        {"name": "id", "offset": 0, "type": "Uint256"},
        {"name": "name_len", "offset": 1, "type": "felt"},
        {"name": "name", "offset": 2, "type": "felt*"},
        {"name": "pool_id", "offset": 3, "type": "PoolId"},
        {"name": "favorite_numbers_triplet", "offset": 4, "type": "felt*"},
    ],
}
user_missing_offset_dict = {
    "type": "struct",
    "name": "User",
    "size": 5,
    "members": [
        {"name": "id", "type": "Uint256"},
        {"name": "name_len", "type": "felt"},
        {"name": "name", "type": "felt*"},
        {"name": "pool_id", "type": "PoolId"},
        {"name": "favorite_numbers_triplet", "type": "felt*"},
    ],
}
user_struct = StructType(
    "User",
    OrderedDict(
        id=uint256_struct,
        name_len=FeltType(),
        name=ArrayType(FeltType()),
        pool_id=pool_id_struct,
        favorite_numbers_triplet=ArrayType(FeltType()),
    ),
)

user_partial_missing_offset_dict = {
    "type": "struct",
    "name": "User",
    "size": 4,
    "members": [
        {"name": "id", "offset": 0, "type": "Uint256"},
        {"name": "name_len", "type": "felt"},
        {"name": "name", "type": "felt*"},
        {"name": "pool_id", "offset": 1, "type": "PoolId"},
    ],
}
user_partial_missing_offset_struct = StructType(
    "User",
    OrderedDict(
        id=uint256_struct,
        pool_id=pool_id_struct,
        name_len=FeltType(),
        name=ArrayType(FeltType()),
    ),
)

user_added_dict = {
    "type": "event",
    "name": "UserAdded",
    "data": [
        {"name": "user", "type": "User"},
    ],
    "keys": [],
}
user_added_event: Abi.Event = Abi.Event(
    "UserAdded",
    OrderedDict(user=user_struct),
)

pool_id_added_dict = {
    "type": "event",
    "name": "PoolIdAdded",
    "data": [
        {"name": "pool_id", "type": "PoolId"},
    ],
    "keys": [],
}
pool_id_added_event: Abi.Event = Abi.Event(
    "PoolIdAdded",
    OrderedDict(pool_id=pool_id_struct),
)

get_user_dict = {
    "type": "function",
    "name": "get_user",
    "inputs": [
        {
            "name": "id",
            "type": "Uint256",
        }
    ],
    "outputs": [{"name": "user", "type": "User"}],
}
get_user_fn = Abi.Function(
    "get_user", OrderedDict(id=uint256_struct), OrderedDict(user=user_struct)
)

delete_pool_dict = {
    "type": "function",
    "name": "delete_pool",
    "inputs": [
        {
            "name": "id",
            "type": "PoolId",
        },
        {"name": "user_id", "type": "Uint256"},
    ],
    "outputs": [],
}
delete_pool_fn = Abi.Function(
    "delete_pool", OrderedDict(id=pool_id_struct, user_id=uint256_struct), OrderedDict()
)

oz_proxy_abi = Abi(
    defined_structures={},
    functions={
        "__default__": Abi.Function(
            name="__default__",
            inputs=OrderedDict(
                [
                    ("selector", FeltType()),
                    ("calldata_size", FeltType()),
                    ("calldata", ArrayType(inner_type=FeltType())),
                ]
            ),
            outputs=OrderedDict(
                [
                    ("retdata_size", FeltType()),
                    ("retdata", ArrayType(inner_type=FeltType())),
                ]
            ),
        )
    },
    constructor=Abi.Function(
        name="constructor",
        inputs=OrderedDict(
            [
                ("implementation_hash", FeltType()),
                ("selector", FeltType()),
                ("calldata_len", FeltType()),
                ("calldata", ArrayType(inner_type=FeltType())),
            ]
        ),
        outputs=OrderedDict(),
    ),
    l1_handler=Abi.Function(
        name="__l1_default__",
        inputs=OrderedDict(
            [
                ("selector", FeltType()),
                ("calldata_size", FeltType()),
                ("calldata", ArrayType(inner_type=FeltType())),
            ]
        ),
        outputs=OrderedDict(),
    ),
    events={
        "Upgraded": Abi.Event(
            name="Upgraded", data=OrderedDict([("implementation", FeltType())])
        ),
        "AdminChanged": Abi.Event(
            name="AdminChanged",
            data=OrderedDict([("previousAdmin", FeltType()), ("newAdmin", FeltType())]),
        ),
    },
)

nested_struct = StructType(
    name="NestedStruct", types=OrderedDict([("value", FeltType())])
)
top_struct = StructType(
    name="TopStruct",
    types=OrderedDict(value=FeltType(), nested_struct=nested_struct),
)
balance_struct_abi = Abi(
    defined_structures={
        "TopStruct": top_struct,
        "NestedStruct": nested_struct,
    },
    functions={
        "increase_balance": Abi.Function(
            name="increase_balance",
            inputs=OrderedDict(
                [
                    ("key", FeltType()),
                    (
                        "amount",
                        top_struct,
                    ),
                ]
            ),
            outputs=OrderedDict(),
        ),
        "get_balance": Abi.Function(
            name="get_balance",
            inputs=OrderedDict([("key", FeltType())]),
            outputs=OrderedDict(value=top_struct),
        ),
    },
    constructor=None,
    l1_handler=None,
    events={
        "increase_balance_called": Abi.Event(
            name="increase_balance_called",
            data=OrderedDict(key=FeltType(), prev_amount=top_struct, amount=top_struct),
        )
    },
)



import pytest

from starknet_py.cairo.felt import encode_shortstring
from starknet_py.constants import STRK_FEE_CONTRACT_ADDRESS
from starknet_py.hash.selector import get_selector_from_name
from starknet_py.net.client_models import Call, TransactionExecutionStatus
from starknet_py.net.models import StarknetChainId
from starknet_py.net.signer.eth_signer import EthSigner
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS


# pylint: disable=line-too-long
# First 2 key pairs are retrieved from the mnemonic "test test test test test test test test test test test junk"
@pytest.mark.parametrize(
    "private_key, expected_public_key",
    [
        (
            "0xAC0974BEC39A17E36BA4A6B4D238FF944BACB478CBED5EFCAE784D7BF4F2FF80",
            0x8318535B54105D4A7AAE60C08FC45F9687181B4FDFC625BD1A753FA7397FED753547F11CA8696646F2F3ACB08E31016AFAC23E630C5D11F59F61FEF57B0D2AA5,
        ),
        (
            0x59C6995E998F97A5A0044966F0945389DC9E86DAE88C7A8412F4603B6B78690D,
            0xBA5734D8F7091719471E7F7ED6B9DF170DC70CC661CA05E688601AD984F068B0D67351E5F06073092499336AB0839EF8A521AFD334E53807205FA2F08EEC74F4,
        ),
        (
            0x525BC68475C0955FAE83869BEEC0996114D4BB27B28B781ED2A20EF23121B8DE,
            0x0178BB97615B49070EEFAD71CB2F159392274404E41DB748D9397147CB25CF597EBFCF2F399E635B72B99B8F76E9080763C65A42C842869815039D912150DDFE,
        ),
    ],
)
def test_pub_key(private_key, expected_public_key):
    eth_signer = EthSigner(
        private_key,
        chain_id=StarknetChainId.MAINNET,
    )
    assert eth_signer.public_key == expected_public_key


@pytest.mark.parametrize(
    "loaded_typed_data, expected_r, expected_s, expected_v",
    [
        (
            "typed_data_rev_0_example.json",
            (0xFF887F391242BB244E9E10D5DA01CB8A, 0x665E69338D4E0772039D4A032B01B07B),
            (0xF84A88E94CABBA842AB4ACCF8ADC0200, 0x61C82A3A2F1A9340620E634BEBECB20B),
            0x1,
        ),
        (
            "typed_data_rev_0_struct_array_example.json",
            (0x936821B8F85EEE978AAF9127814742D2, 0x31BC82F3C096425CFCCBBC08CD06130),
            (0x57E8E66F612B1C612DBFD159BDB8B260, 0x4341F2213DFF450FFF4EA979EA5A48A0),
            0x0,
        ),
        (
            "typed_data_rev_1_example.json",
            (0x47EED1FC641A07E66D699D1F3B1D94E8, 0x9C93AA4098668FECF0CD5E7B37D82A92),
            (0xEE8CC16FF762F7AF5B221AAA75177CC6, 0x4F6CFAB88C2AB909E4B80EBCEE151671),
            0x0,
        ),
    ],
    indirect=["loaded_typed_data"],
)
def test_message_hash(loaded_typed_data, expected_r, expected_s, expected_v):
    eth_signer = EthSigner(
        private_key=0x525BC68475C0955FAE83869BEEC0996114D4BB27B28B781ED2A20EF23121B8DE,
        chain_id=StarknetChainId.MAINNET,
    )
    signed_message = eth_signer.sign_message(
        loaded_typed_data,
        0x65A822FBEE1AE79E898688B5A4282DC79E0042CBED12F6169937FDDB4C26641,
    )

    assert len(signed_message) == 5
    assert signed_message[0] == expected_r[0]
    assert signed_message[1] == expected_r[1]
    assert signed_message[2] == expected_s[0]
    assert signed_message[3] == expected_s[1]
    assert signed_message[4] == expected_v


@pytest.mark.asyncio
async def test_sign_transaction(eth_account, map_contract):
    map_call = Call(
        to_addr=map_contract.address,
        selector=get_selector_from_name("put"),
        calldata=[100, 101],
    )

    signed_tx = await eth_account.sign_invoke_v3(
        calls=map_call, resource_bounds=MAX_RESOURCE_BOUNDS
    )
    calldata = (
        [signed_tx.calculate_hash(StarknetChainId.SEPOLIA)]
        + [len(signed_tx.signature)]
        + signed_tx.signature
    )

    validate_call = Call(
        to_addr=eth_account.address,
        selector=get_selector_from_name("is_valid_signature"),
        calldata=calldata,
    )
    (res,) = await eth_account.client.call_contract(call=validate_call)
    assert res == encode_shortstring("VALID")


@pytest.mark.asyncio
async def test_send_transaction(eth_account):
    call = Call(
        to_addr=int(STRK_FEE_CONTRACT_ADDRESS, 0),
        selector=get_selector_from_name("transfer"),
        calldata=[0x12345, 10000, 0],
    )
    res = await eth_account.execute_v3(calls=call, resource_bounds=MAX_RESOURCE_BOUNDS)

    receipt = await eth_account.client.get_transaction_receipt(res.transaction_hash)

    assert receipt.execution_status == TransactionExecutionStatus.SUCCEEDED



from .array_serializer import ArraySerializer
from .bool_serializer import BoolSerializer
from .byte_array_serializer import ByteArraySerializer
from .cairo_data_serializer import CairoDataSerializer
from .felt_serializer import FeltSerializer
from .named_tuple_serializer import NamedTupleSerializer
from .payload_serializer import PayloadSerializer
from .struct_serializer import StructSerializer
from .tuple_serializer import TupleSerializer
from .uint256_serializer import Uint256Serializer
from .uint_serializer import UintSerializer



from marshmallow import Schema, fields
from marshmallow_oneofschema.one_of_schema import OneOfSchema

from starknet_py.abi.v1.shape import (
    ENUM_ENTRY,
    EVENT_ENTRY,
    FUNCTION_ENTRY,
    STRUCT_ENTRY,
)


class TypeSchema(Schema):
    type = fields.String(data_key="type", required=True)


class TypedParameterSchema(TypeSchema):
    name = fields.String(data_key="name", required=True)


class FunctionBaseSchema(Schema):
    name = fields.String(data_key="name", required=True)
    inputs = fields.List(
        fields.Nested(TypedParameterSchema()), data_key="inputs", required=True
    )
    outputs = fields.List(
        fields.Nested(TypeSchema()), data_key="outputs", required=True
    )
    state_mutability = fields.String(data_key="state_mutability", dump_default=None)


class FunctionAbiEntrySchema(FunctionBaseSchema):
    type = fields.Constant(FUNCTION_ENTRY, data_key="type", required=True)


class EventAbiEntrySchema(Schema):
    type = fields.Constant(EVENT_ENTRY, data_key="type", required=True)
    name = fields.String(data_key="name", required=True)
    inputs = fields.List(
        fields.Nested(TypedParameterSchema()), data_key="inputs", required=True
    )


class StructAbiEntrySchema(Schema):
    type = fields.Constant(STRUCT_ENTRY, data_key="type", required=True)
    name = fields.String(data_key="name", required=True)
    members = fields.List(
        fields.Nested(TypedParameterSchema()), data_key="members", required=True
    )


class EnumAbiEntrySchema(Schema):
    type = fields.Constant(ENUM_ENTRY, data_key="type", required=True)
    name = fields.String(data_key="name", required=True)
    variants = fields.List(
        fields.Nested(TypedParameterSchema(), data_key="variants", required=True)
    )


class ContractAbiEntrySchema(OneOfSchema):
    type_field_remove = False
    type_schemas = {
        FUNCTION_ENTRY: FunctionAbiEntrySchema,
        EVENT_ENTRY: EventAbiEntrySchema,
        STRUCT_ENTRY: StructAbiEntrySchema,
        ENUM_ENTRY: EnumAbiEntrySchema,
    }



# pylint: disable=unused-variable
from starknet_py.devnet_utils.devnet_client import DevnetClient


def test_init():
    # docs-start: init
    client = DevnetClient(node_url="https://your.node.url")
    # docs-end: init



# pylint: disable=redefined-outer-name
from dataclasses import dataclass
from typing import List, Optional, Tuple

import pytest
import pytest_asyncio

from starknet_py.constants import ARGENT_V040_CLASS_HASH
from starknet_py.contract import Contract
from starknet_py.hash.address import compute_address
from starknet_py.net.account.account import Account
from starknet_py.net.account.base_account import BaseAccount
from starknet_py.net.client_models import PriceUnit
from starknet_py.net.full_node_client import FullNodeClient
from starknet_py.net.http_client import HttpMethod, RpcHttpClient
from starknet_py.net.models import AddressRepresentation, StarknetChainId
from starknet_py.net.signer.eth_signer import EthSigner
from starknet_py.net.signer.key_pair import KeyPair
from starknet_py.serialization import Uint256Serializer
from starknet_py.tests.e2e.fixtures.constants import (
    DEVNET_PRE_DEPLOYED_ACCOUNT_ADDRESS,
    DEVNET_PRE_DEPLOYED_ACCOUNT_PRIVATE_KEY,
    MAX_RESOURCE_BOUNDS,
)
from starknet_py.tests.e2e.utils import (
    AccountToBeDeployedDetails,
    _get_random_private_key_unsafe,
    _new_address,
    get_deploy_account_transaction,
    prepay_account,
)


@dataclass
class AccountPrerequisites:
    address: AddressRepresentation
    salt: int
    calldata: list[int]
    key_pair: KeyPair


async def devnet_account_details(
    client: FullNodeClient,
    class_hash: int,
    devnet,
) -> Tuple[str, str]:
    """
    Deploys an Account and adds fee tokens to its balance (only on devnet).
    """

    private_key = _get_random_private_key_unsafe()
    key_pair = KeyPair.from_private_key(private_key)
    salt = 1

    address = compute_address(
        class_hash=class_hash,
        constructor_calldata=[key_pair.public_key],
        salt=salt,
        deployer_address=0,
    )

    await mint_token_on_devnet(devnet, address, int(1e30), PriceUnit.WEI.value)
    await mint_token_on_devnet(devnet, address, int(1e30), PriceUnit.FRI.value)

    deploy_account_tx = await get_deploy_account_transaction(
        address=address,
        key_pair=key_pair,
        salt=salt,
        class_hash=class_hash,
        client=client,
    )

    account = Account(
        address=address,
        client=client,
        key_pair=key_pair,
        chain=StarknetChainId.SEPOLIA,
    )
    res = await account.client.deploy_account(deploy_account_tx)
    await account.client.wait_for_tx(res.transaction_hash)

    return hex(address), hex(key_pair.private_key)


async def mint_token_on_devnet(url: str, address: int, amount: int, unit: str):
    http_client = RpcHttpClient(url)
    await http_client.request(
        http_method=HttpMethod.POST,
        address=f"{url}/mint",
        payload={"address": hex(address), "amount": amount, "unit": unit},
    )


@pytest.fixture(name="account", scope="package")
def full_node_account(client: FullNodeClient) -> BaseAccount:
    """
    Returns a new Account created with FullNodeClient.
    """
    address = DEVNET_PRE_DEPLOYED_ACCOUNT_ADDRESS
    private_key = DEVNET_PRE_DEPLOYED_ACCOUNT_PRIVATE_KEY

    return Account(
        address=address,
        client=client,
        key_pair=KeyPair.from_private_key(int(private_key, 0)),
        chain=StarknetChainId.SEPOLIA,
    )


@dataclass
class AccountToBeDeployedDetailsFactory:
    class_hash: int
    eth_fee_contract: Contract
    strk_fee_contract: Contract

    async def get(
        self,
        *,
        class_hash: Optional[int] = None,
        calldata: Optional[List[int]] = None,
    ) -> AccountToBeDeployedDetails:
        key_pair = KeyPair.from_private_key(_get_random_private_key_unsafe())

        calldata = calldata if calldata is not None else []
        class_hash = class_hash if class_hash is not None else self.class_hash

        if calldata == []:
            calldata = [key_pair.public_key]

        address, salt = _new_address(
            class_hash=class_hash,
            calldata=calldata,
        )

        await prepay_account(
            address=address,
            eth_fee_contract=self.eth_fee_contract,
            strk_fee_contract=self.strk_fee_contract,
        )
        return address, key_pair, salt, class_hash


@pytest_asyncio.fixture(scope="package")
async def deploy_account_details_factory(
    account_with_validate_deploy_class_hash: int,
    eth_fee_contract: Contract,
    strk_fee_contract: Contract,
) -> AccountToBeDeployedDetailsFactory:
    """
    Returns AccountToBeDeployedDetailsFactory.

    The Factory's get() method returns: address, key_pair, salt
    and class_hash of the account with validate deploy.
    Prefunds the address with enough tokens to allow for deployment.
    """
    return AccountToBeDeployedDetailsFactory(
        class_hash=account_with_validate_deploy_class_hash,
        eth_fee_contract=eth_fee_contract,
        strk_fee_contract=strk_fee_contract,
    )


@pytest.fixture(scope="package")
def pre_deployed_account_with_validate_deploy(client) -> BaseAccount:
    """
    Returns an Account pre-deployed on specified network. Used to deploy other accounts.
    """

    address = DEVNET_PRE_DEPLOYED_ACCOUNT_ADDRESS
    private_key = DEVNET_PRE_DEPLOYED_ACCOUNT_PRIVATE_KEY

    return Account(
        address=address,
        client=client,
        key_pair=KeyPair.from_private_key(int(private_key, 16)),
        chain=StarknetChainId.SEPOLIA,
    )


@pytest_asyncio.fixture(scope="function")
async def argent_account_v040_data(
    eth_fee_contract: Contract,
    strk_fee_contract: Contract,
) -> AccountPrerequisites:
    key_pair = KeyPair.from_private_key(_get_random_private_key_unsafe())

    # Based on ABI definition documentation
    constructor_calldata = [
        0,
        key_pair.public_key,
        1,
    ]

    address, salt = _new_address(ARGENT_V040_CLASS_HASH, constructor_calldata)

    await prepay_account(
        address=address,
        eth_fee_contract=eth_fee_contract,
        strk_fee_contract=strk_fee_contract,
    )

    return AccountPrerequisites(address, salt, constructor_calldata, key_pair)


@pytest_asyncio.fixture(scope="package")
async def argent_account_v040(
    eth_fee_contract: Contract,
    strk_fee_contract: Contract,
    client,
) -> BaseAccount:
    key_pair = KeyPair.from_private_key(_get_random_private_key_unsafe())

    # Based on ABI definition documentation
    constructor_calldata = [
        0,
        key_pair.public_key,
        1,
    ]

    address, salt = _new_address(ARGENT_V040_CLASS_HASH, constructor_calldata)

    await prepay_account(
        address=address,
        eth_fee_contract=eth_fee_contract,
        strk_fee_contract=strk_fee_contract,
    )

    deploy_result = await Account.deploy_account_v3(
        address=address,
        class_hash=ARGENT_V040_CLASS_HASH,
        salt=salt,
        key_pair=key_pair,
        client=client,
        constructor_calldata=constructor_calldata,
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )

    await deploy_result.wait_for_acceptance()
    return deploy_result.account


@pytest_asyncio.fixture(scope="package")
async def eth_account(
    eth_fee_contract: Contract,
    strk_fee_contract: Contract,
    client,
    eth_account_class_hash,
) -> BaseAccount:
    private_key = _get_random_private_key_unsafe()

    signer = EthSigner(private_key, chain_id=StarknetChainId.SEPOLIA)

    # Manually serialize u512 into felt array
    serializer = Uint256Serializer()
    public_key_bytes = signer.public_key.to_bytes(64, byteorder="big")
    constructor_calldata = serializer.serialize(
        int.from_bytes(public_key_bytes[:32], byteorder="big")
    ) + serializer.serialize(int.from_bytes(public_key_bytes[32:], byteorder="big"))

    address, salt = _new_address(eth_account_class_hash, constructor_calldata)

    await prepay_account(
        address=address,
        eth_fee_contract=eth_fee_contract,
        strk_fee_contract=strk_fee_contract,
    )

    account = Account(
        address=address,
        client=client,
        signer=signer,
        chain=StarknetChainId.SEPOLIA,
    )

    deploy_account_tx = await account.sign_deploy_account_v3(
        class_hash=eth_account_class_hash,
        contract_address_salt=salt,
        constructor_calldata=constructor_calldata,
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )

    await client.deploy_account(deploy_account_tx)

    return account



# pylint: disable=unused-variable
import pytest

from starknet_py.contract import Contract, ContractFunction
from starknet_py.net.client_models import ResourceBounds, ResourceBoundsMapping


def test_prepare_invoke_v3(map_contract: Contract):
    # docs-start: prepare_invoke_v3
    prepared_function_call = map_contract.functions["put"].prepare_invoke_v3(
        key=10, value=20
    )
    # docs-end: prepare_invoke_v3


@pytest.mark.asyncio
async def test_call(map_contract: Contract):
    # docs-start: call
    call_result = map_contract.functions["get"].call(key=10)
    # or when call has to be done at specific block
    call_result = map_contract.functions["get"].call(key=10, block_number="latest")
    # docs-end: call


def test_invoke_v3(map_contract: Contract):
    # docs-start: invoke_v3
    resource_bounds = ResourceBoundsMapping(
        l1_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
        l2_gas=ResourceBounds(max_amount=int(1e9), max_price_per_unit=int(1e17)),
        l1_data_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
    )
    invoke_result = map_contract.functions["put"].invoke_v3(
        key=10,
        value=20,
        resource_bounds=resource_bounds,
    )
    # docs-end: invoke_v3


def test_get_selector():
    # docs-start: get_selector
    selector = ContractFunction.get_selector("get")
    # docs-end: get_selector



import pytest

import starknet_py.tests.e2e.fixtures.abi_v1_structures as fixtures
from starknet_py.abi.v1.parser import AbiParser, AbiParsingError
from starknet_py.cairo.v1.type_parser import UnknownCairoTypeError


def test_parsing_types_abi():
    # Even though user depend on pool id and uint256 it is defined first. Parser has to consider those cases
    abi = AbiParser(
        [
            fixtures.user_dict,
            fixtures.pool_id_dict,
            fixtures.user_added_dict,
            fixtures.pool_id_added_dict,
            fixtures.get_user_dict,
            fixtures.delete_pool_dict,
        ]
    ).parse()

    assert abi.defined_structures == {
        "PoolId": fixtures.pool_id_struct,
        "User": fixtures.user_struct,
        **fixtures.core_structures,
    }
    assert abi.events == {
        "UserAdded": fixtures.user_added_event,
        "PoolIdAdded": fixtures.pool_id_added_event,
    }
    assert abi.functions == {
        "get_user": fixtures.get_user_fn,
        "delete_pool": fixtures.delete_pool_fn,
    }


def test_parsing_types_abi2():
    abi = AbiParser(
        [
            fixtures.foo_external_dict,
            fixtures.foo_event_dict,
            fixtures.foo_view_dict,
            fixtures.my_enum_dict,
            fixtures.my_struct_dict,
        ]
    ).parse()

    assert abi.defined_structures == {
        "test::MyStruct::<core::integer::u256>": fixtures.my_struct,
        **fixtures.core_structures,
    }
    assert abi.defined_enums == {
        "test::MyEnum::<core::integer::u128>": fixtures.my_enum,
    }
    assert abi.events == {
        "foo_event": fixtures.foo_event,
    }
    assert abi.functions == {
        "foo_external": fixtures.foo_external,
        "foo_view": fixtures.foo_view,
    }


def test_self_cycle():
    self_referencing_struct = {
        "type": "struct",
        "name": "Infinite",
        "members": [
            {"name": "value", "type": "Infinite"},
        ],
    }
    with pytest.raises(
        AbiParsingError,
        match="Circular reference detected",
    ):
        AbiParser([self_referencing_struct]).parse()


def test_bigger_cycle():
    # first -> seconds -> third -> first...
    first = {
        "type": "struct",
        "name": "First",
        "members": [{"name": "value", "type": "Second"}],
    }
    second = {
        "type": "struct",
        "name": "Second",
        "members": [{"name": "value", "type": "Third"}],
    }
    third = {
        "type": "struct",
        "name": "Third",
        "members": [{"name": "value", "type": "First"}],
    }
    with pytest.raises(
        AbiParsingError,
        match="Circular reference detected",
    ):
        AbiParser([first, second, third]).parse()


def test_duplicated_structure():
    with pytest.raises(
        AbiParsingError,
        match="Name 'User' was used more than once in defined structures",
    ):
        AbiParser(
            [fixtures.user_dict, fixtures.pool_id_dict, fixtures.user_dict]
        ).parse()


def test_duplicated_function():
    with pytest.raises(
        AbiParsingError,
        match="Name 'get_user' was used more than once in defined functions",
    ):
        AbiParser(
            [
                fixtures.get_user_dict,
                fixtures.delete_pool_dict,
                fixtures.get_user_dict,
                fixtures.delete_pool_dict,
            ]
        ).parse()


def test_duplicated_event():
    with pytest.raises(
        AbiParsingError,
        match="Name 'UserAdded' was used more than once in defined events",
    ):
        AbiParser(
            [
                fixtures.user_added_dict,
                fixtures.delete_pool_dict,
                fixtures.user_added_dict,
            ]
        ).parse()


def test_duplicated_type_members():
    type_dict = {
        "type": "struct",
        "name": "Record",
        "members": [
            {"name": "name", "type": "core::felt252"},
            {"name": "value", "type": "core::felt252"},
            {"name": "id", "type": "core::felt252"},
            {"name": "value", "type": "core::felt252"},
        ],
    }
    with pytest.raises(
        AbiParsingError,
        match="Name 'value' was used more than once in members of structure 'Record'",
    ):
        AbiParser([type_dict]).parse()


@pytest.mark.parametrize(
    "missing_name, input_dict",
    [
        # Type
        ("PoolId", fixtures.user_dict),
        # Function
        ("User", fixtures.get_user_dict),
        # Event
        ("User", fixtures.user_added_dict),
    ],
)
def test_missing_type_used(missing_name, input_dict):
    with pytest.raises(
        UnknownCairoTypeError, match=f"Type '{missing_name}' is not defined.*"
    ):
        AbiParser([input_dict]).parse()



from typing import Optional


class WebsocketClientError(Exception):
    """
    Base class for all errors raised while attempting to communicate with Starknet through WebsocketClient.
    """

    def __init__(
        self, message: str, code: Optional[str] = None, data: Optional[str] = None
    ):
        self.code = code
        self.data = data
        self.message = (
            f"WebsocketClient failed{f' with code {code}' if code is not None else ''}. "
            f"Message: {message}.{f' Data: {data}' if data is not None else ''}"
        )

        super().__init__(self.message)



from typing import List

import pytest

from starknet_py.devnet_utils.devnet_client_models import PredeployedAccount
from starknet_py.net.client_models import PriceUnit


@pytest.mark.asyncio
async def test_mint(devnet_client, account):
    amount = 1000

    balance_before_mint = await account.get_balance()
    await devnet_client.mint(account.address, amount, PriceUnit.FRI)
    await devnet_client.mint(account.address, amount, "fri")
    await devnet_client.mint(account.address, amount, "FRI")
    await devnet_client.mint(account.address, amount)
    balance_after_mint = await account.get_balance()

    assert balance_after_mint == balance_before_mint + 4 * amount


@pytest.mark.asyncio
async def test_create_blocks(devnet_client):
    block_hash = await devnet_client.create_block()
    assert block_hash is not None


@pytest.mark.asyncio
async def test_abort_blocks(devnet_client):
    block_hash = await devnet_client.create_block()
    for _ in range(5):
        await devnet_client.create_block()

    aborted_blocks = await devnet_client.abort_block(block_hash=block_hash)
    assert len(aborted_blocks) == 6


@pytest.mark.asyncio
async def test_predeployed_accounts(devnet_client):
    accounts = await devnet_client.get_predeployed_accounts()

    isinstance(accounts, List)
    assert len(accounts) > 0
    isinstance(accounts[0], PredeployedAccount)



from typing import Iterable, TypeVar, Union

T = TypeVar("T")


def ensure_iterable(value: Union[T, Iterable[T]]) -> Iterable[T]:
    if isinstance(value, Iterable):
        return value
    return [value]



import pytest

from starknet_py.net.client_models import TransactionExecutionStatus
from starknet_py.net.models.transaction import DeclareV3
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS


@pytest.mark.asyncio
async def test_declare_v3_tx(account, abi_types_compiled_contract_and_class_hash):
    declare_tx = await account.sign_declare_v3(
        compiled_contract=abi_types_compiled_contract_and_class_hash[0],
        compiled_class_hash=abi_types_compiled_contract_and_class_hash[1],
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )
    assert isinstance(declare_tx, DeclareV3)

    result = await account.client.declare(declare_tx)
    tx_receipt = await account.client.wait_for_tx(tx_hash=result.transaction_hash)

    assert tx_receipt.execution_status == TransactionExecutionStatus.SUCCEEDED



import pytest

from starknet_py.net.client_models import (
    ResourceBoundsMapping,
    TransactionFinalityStatus,
)


@pytest.mark.asyncio
async def test_account_outside_execution_any_caller(
    account,
    argent_account_v040,
    map_contract,
):
    # pylint: disable=import-outside-toplevel,too-many-locals

    # docs: start
    import datetime

    from starknet_py.constants import ANY_CALLER
    from starknet_py.hash.selector import get_selector_from_name
    from starknet_py.net.client_models import (
        Call,
        OutsideExecutionTimeBounds,
        ResourceBounds,
    )

    # Create a call to put value 100 at key 1.
    put_call = Call(
        to_addr=map_contract.address,
        selector=get_selector_from_name("put"),
        calldata=[1, 100],
    )

    # Create an outside execution call. This call can now be executed by
    # the specified caller. In this case, anyone will be able to execute it.

    # Note that signing account does not need to have any funds to sign the transaction.
    call = await argent_account_v040.sign_outside_execution_call(
        calls=[
            put_call,
        ],
        # The transaction can be executed in specified timeframe.
        execution_time_bounds=OutsideExecutionTimeBounds(
            execute_after=datetime.datetime.now() - datetime.timedelta(hours=1),
            execute_before=datetime.datetime.now() + datetime.timedelta(hours=1),
        ),
        # Use ANY_CALLER, a special constant that allows anyone to execute the call.
        caller=ANY_CALLER,
    )

    # Now, if you're in specified timeframe, you can perform the outside execution by another account.
    tx = await account.execute_v3(
        calls=[call],
        resource_bounds=ResourceBoundsMapping(
            l1_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
            l2_gas=ResourceBounds(max_amount=int(1e9), max_price_per_unit=int(1e17)),
            l1_data_gas=ResourceBounds(
                max_amount=int(1e5), max_price_per_unit=int(1e13)
            ),
        ),
    )
    await account.client.wait_for_tx(tx.transaction_hash)

    # docs: end

    receipt = await account.client.get_transaction_receipt(tx_hash=tx.transaction_hash)

    assert receipt.finality_status == TransactionFinalityStatus.ACCEPTED_ON_L2



import dataclasses
import sys
from unittest.mock import AsyncMock, patch

import pytest

from starknet_py.common import create_casm_class
from starknet_py.contract import Contract
from starknet_py.hash.address import compute_address
from starknet_py.hash.casm_class_hash import compute_casm_class_hash
from starknet_py.hash.selector import get_selector_from_name
from starknet_py.hash.storage import get_storage_var_address
from starknet_py.net.account.account import Account
from starknet_py.net.client_errors import ClientError
from starknet_py.net.client_models import (
    BlockHashAndNumber,
    Call,
    DeclareTransactionTrace,
    DeclareTransactionV3,
    DeployAccountTransactionTrace,
    InvokeTransactionTrace,
    SierraContractClass,
    SimulatedTransaction,
    SyncStatus,
    TransactionType,
)
from starknet_py.net.full_node_client import _to_rpc_felt
from starknet_py.net.models import StarknetChainId
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS
from starknet_py.tests.e2e.fixtures.misc import ContractVersion, load_contract
from starknet_py.tests.e2e.utils import create_empty_block


def _parse_event_name(event: str) -> str:
    return _to_rpc_felt(get_selector_from_name(event))


FUNCTION_ONE_NAME = "put"
EVENT_ONE_PARSED_NAME = _parse_event_name("PutCalled")
FUNCTION_TWO_NAME = "another_put"
EVENT_TWO_PARSED_NAME = _parse_event_name("AnotherPutCalled")


@pytest.mark.run_on_devnet
@pytest.mark.asyncio
async def test_node_get_declare_transaction_by_block_number_and_index(
    declare_transaction_hash, block_with_declare_number, client, class_hash
):
    tx = await client.get_transaction_by_block_id(
        block_number=block_with_declare_number, index=0
    )

    assert isinstance(tx, DeclareTransactionV3)
    assert tx.hash == declare_transaction_hash
    assert tx.class_hash == class_hash
    assert tx.version == 3


@pytest.mark.run_on_devnet
@pytest.mark.asyncio
async def test_get_class_at(
    client, contract_address, hello_starknet_deploy_transaction_address
):
    declared_contract = await client.get_class_at(
        contract_address=contract_address, block_hash="latest"
    )

    assert isinstance(declared_contract, SierraContractClass)
    assert declared_contract.sierra_program != {}
    assert declared_contract.entry_points_by_type is not None
    assert declared_contract.abi is not None

    declared_contract = await client.get_class_at(
        contract_address=hello_starknet_deploy_transaction_address, block_hash="latest"
    )
    assert isinstance(declared_contract, SierraContractClass)
    assert declared_contract.sierra_program != {}
    assert declared_contract.entry_points_by_type is not None
    assert declared_contract.abi is not None


@pytest.mark.run_on_devnet
@pytest.mark.asyncio
async def test_get_class_at_throws_on_wrong_address(client):
    with pytest.raises(
        ClientError, match="Client failed with code 20. Message: Contract not found."
    ):
        await client.get_class_at(contract_address=0, block_hash="latest")


@pytest.mark.run_on_devnet
@pytest.mark.asyncio
async def test_block_transaction_count(client):
    latest_block = await client.get_block("latest")

    for block_number in range(1, latest_block.block_number + 1):
        transaction_count = await client.get_block_transaction_count(
            block_number=block_number
        )

        assert transaction_count == 1


@pytest.mark.asyncio
async def test_method_raises_on_both_block_hash_and_number(client):
    with pytest.raises(
        ValueError,
        match="Arguments block_hash and block_number are mutually exclusive.",
    ):
        await client.get_block(block_number=0, block_hash="0x0")


@pytest.mark.asyncio
async def test_get_transaction_receipt_deploy_account(
    client, deploy_account_details_factory
):
    address, key_pair, salt, class_hash = await deploy_account_details_factory.get()
    deploy_result = await Account.deploy_account_v3(
        address=address,
        class_hash=class_hash,
        salt=salt,
        key_pair=key_pair,
        client=client,
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )
    await deploy_result.wait_for_acceptance()

    receipt = await client.get_transaction_receipt(tx_hash=deploy_result.hash)
    assert receipt.type == TransactionType.DEPLOY_ACCOUNT
    assert receipt.contract_address == deploy_result.account.address


@pytest.mark.run_on_devnet
@pytest.mark.asyncio
async def test_get_storage_at_incorrect_address_full_node_client(client):
    with pytest.raises(ClientError, match="Contract not found"):
        await client.get_storage_at(
            contract_address=0x1111,
            key=get_storage_var_address("balance"),
            block_hash="latest",
        )


@pytest.mark.skipif(
    "--contract_dir=v1" in sys.argv,
    reason="Contract exists only in v2 directory",
)
@pytest.mark.skip("TODO(#1659)")
@pytest.mark.run_on_devnet
@pytest.mark.asyncio
async def test_get_events_without_following_continuation_token(
    client,
    simple_storage_with_event_contract: Contract,
):
    for i in range(4):
        await simple_storage_with_event_contract.functions[FUNCTION_ONE_NAME].invoke_v3(
            i, i, auto_estimate=True
        )

    chunk_size = 3
    events_response = await client.get_events(
        from_block_number=0,
        to_block_hash="latest",
        address=simple_storage_with_event_contract.address,
        keys=[[EVENT_ONE_PARSED_NAME]],
        follow_continuation_token=False,
        chunk_size=chunk_size,
    )

    assert len(events_response.events) == chunk_size
    assert events_response.continuation_token is not None


@pytest.mark.skipif(
    "--contract_dir=v1" in sys.argv,
    reason="Contract exists only in v2 directory",
)
@pytest.mark.skip("TODO(#1659)")
@pytest.mark.run_on_devnet
@pytest.mark.asyncio
async def test_get_events_follow_continuation_token(
    client,
    simple_storage_with_event_contract: Contract,
):
    total_invokes = 2
    for i in range(total_invokes):
        await simple_storage_with_event_contract.functions[FUNCTION_ONE_NAME].invoke_v3(
            i, i + 1, auto_estimate=True
        )

    events_response = await client.get_events(
        from_block_number=0,
        to_block_hash="latest",
        address=simple_storage_with_event_contract.address,
        keys=[[EVENT_ONE_PARSED_NAME]],
        follow_continuation_token=True,
        chunk_size=1,
    )

    assert len(events_response.events) == total_invokes
    assert events_response.continuation_token is None


@pytest.mark.skipif(
    "--contract_dir=v1" in sys.argv,
    reason="Contract exists only in v2 directory",
)
@pytest.mark.run_on_devnet
@pytest.mark.asyncio
async def test_get_events_nonexistent_event_name(
    client,
    simple_storage_with_event_contract: Contract,
):
    await simple_storage_with_event_contract.functions[FUNCTION_ONE_NAME].invoke_v3(
        1, 1, auto_estimate=True
    )

    events_response = await client.get_events(
        from_block_number=0,
        to_block_hash="latest",
        address=simple_storage_with_event_contract.address,
        keys=[[_parse_event_name("nonexistent_event")]],
        follow_continuation_token=False,
        chunk_size=3,
    )

    assert len(events_response.events) == 0
    assert events_response.continuation_token is None


@pytest.mark.skipif(
    "--contract_dir=v1" in sys.argv,
    reason="Contract exists only in v2 directory",
)
@pytest.mark.skip("TODO(#1659)")
@pytest.mark.run_on_devnet
@pytest.mark.asyncio
async def test_get_events_with_two_events(
    client,
    simple_storage_with_event_contract: Contract,
):
    invokes_of_one = 1
    invokes_of_two = 2
    invokes_of_all = invokes_of_one + invokes_of_two
    await simple_storage_with_event_contract.functions[FUNCTION_ONE_NAME].invoke_v3(
        1, 2, auto_estimate=True
    )
    for i in range(invokes_of_two):
        await simple_storage_with_event_contract.functions[FUNCTION_TWO_NAME].invoke_v3(
            i, i + 1, auto_estimate=True
        )

    event_one_events_response = await client.get_events(
        from_block_number=0,
        to_block_hash="latest",
        address=simple_storage_with_event_contract.address,
        keys=[[EVENT_ONE_PARSED_NAME]],
        follow_continuation_token=True,
    )
    event_two_events_response = await client.get_events(
        from_block_number=0,
        to_block_hash="latest",
        address=simple_storage_with_event_contract.address,
        keys=[[EVENT_TWO_PARSED_NAME]],
        follow_continuation_token=True,
    )
    event_one_two_events_response = await client.get_events(
        from_block_number=0,
        to_block_hash="latest",
        address=simple_storage_with_event_contract.address,
        keys=[[int(EVENT_ONE_PARSED_NAME, 0), int(EVENT_TWO_PARSED_NAME, 0)]],
        follow_continuation_token=True,
    )

    assert len(event_one_events_response.events) == invokes_of_one
    assert event_one_events_response.continuation_token is None

    assert len(event_two_events_response.events) == invokes_of_two
    assert event_two_events_response.continuation_token is None

    assert len(event_one_two_events_response.events) == invokes_of_all
    assert event_one_two_events_response.continuation_token is None


@pytest.mark.skipif(
    "--contract_dir=v1" in sys.argv,
    reason="Contract exists only in v2 directory",
)
@pytest.mark.skip("TODO(#1659)")
@pytest.mark.run_on_devnet
@pytest.mark.asyncio
async def test_get_events_start_from_continuation_token(
    client,
    simple_storage_with_event_contract: Contract,
):
    for i in range(5):
        await simple_storage_with_event_contract.functions[FUNCTION_ONE_NAME].invoke_v3(
            i, i + 1, auto_estimate=True
        )

    chunk_size = 2
    continuation_token = "1"
    events_response = await client.get_events(
        from_block_number=0,
        to_block_hash="latest",
        address=simple_storage_with_event_contract.address,
        keys=[[EVENT_ONE_PARSED_NAME]],
        continuation_token=continuation_token,
        chunk_size=chunk_size,
    )
    expected_continuation_token = str(int(continuation_token) + 1)

    assert len(events_response.events) == chunk_size
    assert events_response.continuation_token == expected_continuation_token


@pytest.mark.skipif(
    "--contract_dir=v1" in sys.argv,
    reason="Contract exists only in v2 directory",
)
@pytest.mark.skip("TODO(#1659)")
@pytest.mark.run_on_devnet
@pytest.mark.asyncio
async def test_get_events_no_params(
    client,
    simple_storage_with_event_contract: Contract,
):
    default_chunk_size = 1
    for i in range(3):
        await simple_storage_with_event_contract.functions[FUNCTION_ONE_NAME].invoke_v3(
            i, i + 1, auto_estimate=True
        )
        await simple_storage_with_event_contract.functions[FUNCTION_TWO_NAME].invoke_v3(
            i, i + 1, auto_estimate=True
        )
    events_response = await client.get_events()

    assert len(events_response.events) == default_chunk_size


@pytest.mark.skipif(
    "--contract_dir=v1" in sys.argv,
    reason="Contract exists only in v2 directory",
)
@pytest.mark.run_on_devnet
@pytest.mark.asyncio
async def test_get_events_nonexistent_starting_block(
    client,
    simple_storage_with_event_contract: Contract,
):
    with pytest.raises(ClientError):
        await client.get_events(
            from_block_number=10000,
            to_block_hash="latest",
            address=simple_storage_with_event_contract.address,
            keys=[[EVENT_ONE_PARSED_NAME]],
            follow_continuation_token=False,
            chunk_size=1,
        )


@pytest.mark.asyncio
async def test_get_block_number(client):
    # pylint: disable=protected-access
    await create_empty_block(client._client)

    block_number = await client.get_block_number()

    # pylint: disable=protected-access
    await create_empty_block(client._client)

    new_block_number = await client.get_block_number()
    assert new_block_number == block_number + 1


@pytest.mark.asyncio
async def test_get_block_hash_and_number(client):
    # pylint: disable=protected-access
    await create_empty_block(client._client)

    block_hash_and_number = await client.get_block_hash_and_number()

    assert isinstance(block_hash_and_number, BlockHashAndNumber)

    # pylint: disable=protected-access
    await create_empty_block(client._client)

    new_block_hash_and_number = await client.get_block_hash_and_number()

    assert (
        new_block_hash_and_number.block_number == block_hash_and_number.block_number + 1
    )
    assert new_block_hash_and_number.block_hash > 0


@pytest.mark.asyncio
async def test_get_chain_id(client):
    chain_id = await client.get_chain_id()

    assert chain_id == hex(StarknetChainId.SEPOLIA.value)


@pytest.mark.asyncio
async def test_get_syncing_status_false(client):
    sync_status = await client.get_syncing_status()

    assert sync_status is False


@pytest.mark.asyncio
async def test_get_syncing_status(client):
    with patch(
        "starknet_py.net.http_client.RpcHttpClient.call", AsyncMock()
    ) as mocked_status:
        mocked_status.return_value = {
            "starting_block_num": "0xc8023",
            "current_block_num": "0xc9773",
            "highest_block_num": "0xc9773",
            "starting_block_hash": "0x60be3a55621597c15a53a1f83e977ca5e52e775ab2ebf572d2ebd6a8168fc88",
            "current_block_hash": "0x79abcb48e71524ad2e123624b0ee3d5f69f99759a23441f6f363794d0687a66",
            "highest_block_hash": "0x79abcb48e71524ad2e123624b0ee3d5f69f99759a23441f6f363794d0687a66",
        }

        sync_status = await client.get_syncing_status()

    assert isinstance(sync_status, SyncStatus)


# ---------------------------- Trace API tests ----------------------------


@pytest.mark.asyncio
@pytest.mark.skip
async def test_simulate_transactions_skip_validate(account, deployed_balance_contract):
    assert isinstance(deployed_balance_contract, Contract)
    call = Call(
        to_addr=deployed_balance_contract.address,
        selector=get_selector_from_name("increase_balance"),
        calldata=[0x10],
    )

    invoke_tx = await account.sign_invoke_v3(calls=call, auto_estimate=True)
    invoke_tx = dataclasses.replace(invoke_tx, signature=[])

    simulated_txs = await account.client.simulate_transactions(
        transactions=[invoke_tx], skip_validate=True, block_number="latest"
    )
    assert simulated_txs[0].transaction_trace.validate_invocation is None

    with pytest.raises(ClientError, match="Contract error"):
        await account.client.simulate_transactions(
            transactions=[invoke_tx], block_number="latest"
        )


@pytest.mark.asyncio
@pytest.mark.skip(
    "TODO(#1560): There is no `l1_data_gas` in `execution_resources` from devnet"
)
async def test_simulate_transactions_skip_fee_charge(
    account, deployed_balance_contract
):
    assert isinstance(deployed_balance_contract, Contract)
    call = Call(
        to_addr=deployed_balance_contract.address,
        selector=get_selector_from_name("increase_balance"),
        calldata=[0x10],
    )

    invoke_tx = await account.sign_invoke_v3(calls=call, auto_estimate=True)

    simulated_txs = await account.client.simulate_transactions(
        transactions=[invoke_tx], skip_fee_charge=True, block_number="latest"
    )
    assert simulated_txs is not None


@pytest.mark.skip(reason="TODO(#1659)")
@pytest.mark.asyncio
async def test_simulate_transactions_invoke(account, deployed_balance_contract):
    assert isinstance(deployed_balance_contract, Contract)
    call = Call(
        to_addr=deployed_balance_contract.address,
        selector=get_selector_from_name("increase_balance"),
        calldata=[0x10],
    )

    invoke_tx = await account.sign_invoke_v3(calls=call, auto_estimate=True)
    simulated_txs = await account.client.simulate_transactions(
        transactions=[invoke_tx], block_number="latest"
    )

    assert isinstance(simulated_txs[0], SimulatedTransaction)
    assert isinstance(simulated_txs[0].transaction_trace, InvokeTransactionTrace)
    assert simulated_txs[0].transaction_trace.execute_invocation is not None
    assert simulated_txs[0].transaction_trace.execution_resources is not None

    invoke_tx = await account.sign_invoke_v3(calls=[call, call], auto_estimate=True)
    simulated_txs = await account.client.simulate_transactions(
        transactions=[invoke_tx], block_number="latest"
    )

    assert isinstance(simulated_txs[0].transaction_trace, InvokeTransactionTrace)
    assert simulated_txs[0].transaction_trace.validate_invocation is not None
    assert simulated_txs[0].transaction_trace.execute_invocation is not None
    assert simulated_txs[0].transaction_trace.execution_resources is not None


@pytest.mark.skip(reason="TODO(#1659)")
@pytest.mark.asyncio
async def test_simulate_transactions_two_txs(account, deployed_balance_contract):
    assert isinstance(deployed_balance_contract, Contract)
    call = Call(
        to_addr=deployed_balance_contract.address,
        selector=get_selector_from_name("increase_balance"),
        calldata=[0x10],
    )
    invoke_tx = await account.sign_invoke_v3(calls=call, auto_estimate=True)

    contract = load_contract(
        contract_name="TestContractDeclare", version=ContractVersion.V1
    )

    casm_class = create_casm_class(contract["casm"])
    casm_class_hash = compute_casm_class_hash(casm_class)

    declare_v3_tx = await account.sign_declare_v3(
        compiled_contract=contract["sierra"],
        compiled_class_hash=casm_class_hash,
        # because raw calls do not increment nonce, it needs to be done manually
        nonce=invoke_tx.nonce + 1,
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )

    simulated_txs = await account.client.simulate_transactions(
        transactions=[invoke_tx, declare_v3_tx], block_number="latest"
    )

    assert isinstance(simulated_txs[0].transaction_trace, InvokeTransactionTrace)
    assert simulated_txs[0].fee_estimation.overall_fee > 0
    assert simulated_txs[0].transaction_trace.validate_invocation is not None
    assert simulated_txs[0].transaction_trace.execute_invocation is not None
    assert simulated_txs[0].transaction_trace.execution_resources is not None

    assert isinstance(simulated_txs[1].transaction_trace, DeclareTransactionTrace)
    assert simulated_txs[1].fee_estimation.overall_fee > 0
    assert simulated_txs[1].transaction_trace.validate_invocation is not None
    assert simulated_txs[1].transaction_trace.execution_resources is not None


@pytest.mark.asyncio
@pytest.mark.skip("TODO(#1560)")
async def test_simulate_transactions_deploy_account(
    client, deploy_account_details_factory
):
    address, key_pair, salt, class_hash = await deploy_account_details_factory.get()
    address = compute_address(
        salt=salt,
        class_hash=class_hash,
        constructor_calldata=[key_pair.public_key],
        deployer_address=0,
    )
    account = Account(
        address=address,
        client=client,
        key_pair=key_pair,
        chain=StarknetChainId.SEPOLIA,
    )
    deploy_account_tx = await account.sign_deploy_account_v3(
        class_hash=class_hash,
        contract_address_salt=salt,
        constructor_calldata=[key_pair.public_key],
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )

    simulated_txs = await client.simulate_transactions(
        transactions=[deploy_account_tx], block_number="latest"
    )

    assert isinstance(simulated_txs[0].transaction_trace, DeployAccountTransactionTrace)
    assert simulated_txs[0].fee_estimation.overall_fee > 0
    assert simulated_txs[0].transaction_trace.constructor_invocation is not None
    assert simulated_txs[0].transaction_trace.execution_resources is not None



import os
from pathlib import Path

from dotenv import load_dotenv

from starknet_py.net.client_models import ResourceBounds, ResourceBoundsMapping

load_dotenv(dotenv_path=Path(os.path.dirname(__file__)) / "../test-variables.env")


def _get_env_or_throw(env_name: str) -> str:
    env = os.getenv(key=env_name)
    if env is None:
        raise ValueError(
            f"{env_name} environmental variable is not set. "
            f"Update it manually or set it in `starknet_py/tests/e2e/test-variables.env` file. "
            f"More info here: https://starknetpy.readthedocs.io/en/latest/development.html#setup"
        )
    return env


def _get_env_lambda(env_name):
    return lambda: _get_env_or_throw(env_name)


# -------------------------------- SEPOLIA TESTNET -------------------------------------

SEPOLIA_ACCOUNT_PRIVATE_KEY = _get_env_lambda("SEPOLIA_ACCOUNT_PRIVATE_KEY")

SEPOLIA_ACCOUNT_ADDRESS = _get_env_lambda("SEPOLIA_ACCOUNT_ADDRESS")

SEPOLIA_RPC_URL = _get_env_lambda("SEPOLIA_RPC_URL")

EMPTY_CONTRACT_ADDRESS_SEPOLIA = (
    "0x06524771cb912945bf2db355b5a12355ca2e2ff05e15ee35366336a602293f2d"
)

# -----------------------------------------------------------------------------

DEVNET_PRE_DEPLOYED_ACCOUNT_ADDRESS = (
    "0x260a8311b4f1092db620b923e8d7d20e76dedcc615fb4b6fdf28315b81de201"
)
DEVNET_PRE_DEPLOYED_ACCOUNT_PRIVATE_KEY = "0xc10662b7b247c7cecf7e8a30726cff12"

STRK_CLASS_HASH = "0x04ad3c1dc8413453db314497945b6903e1c766495a1e60492d44da9c2a986e4b"

MAX_RESOURCE_BOUNDS_L1 = ResourceBounds(
    max_amount=int(1e5), max_price_per_unit=int(1e13)
)

MAX_RESOURCE_BOUNDS = ResourceBoundsMapping(
    l1_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
    l2_gas=ResourceBounds(max_amount=int(1e10), max_price_per_unit=int(1e20)),
    l1_data_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
)

MAX_RESOURCE_BOUNDS_SEPOLIA = ResourceBoundsMapping(
    l1_gas=ResourceBounds(max_amount=int(1e4), max_price_per_unit=int(1e15)),
    l2_gas=ResourceBounds(max_amount=int(1e6), max_price_per_unit=int(1e10)),
    l1_data_gas=ResourceBounds(max_amount=int(1e4), max_price_per_unit=int(1e15)),
)

MOCK_DIR = Path(os.path.dirname(__file__)) / "../mock"
TYPED_DATA_DIR = MOCK_DIR / "typed_data"
CONTRACTS_DIR = MOCK_DIR / "contracts"
CONTRACTS_COMPILED_V0_DIR = MOCK_DIR / "contracts_compiled"
CAIRO_0_CONTRACTS_ABI_DIR = MOCK_DIR / "cairo_0_contracts_abi"


# PRECOMPILED_CONTRACTS are contracts compiled with various Sierras
# They are mainly used to verify if we compute class_hash for older Sierras correctly
PRECOMPILED_CONTRACTS_DIR = MOCK_DIR / "precompiled_contracts"

CONTRACTS_V2_COMPILED = MOCK_DIR / "contracts_v2/target/dev"
CONTRACTS_V2_ARTIFACTS_MAP = (
    CONTRACTS_V2_COMPILED / "contracts_v2.starknet_artifacts.json"
)

CONTRACTS_V1_COMPILED = MOCK_DIR / "contracts_v1/target/dev"
CONTRACTS_V1_ARTIFACTS_MAP = (
    CONTRACTS_V1_COMPILED / "contracts_v1.starknet_artifacts.json"
)



# pylint: disable=import-outside-toplevel, pointless-string-statement, unbalanced-tuple-unpacking
import json

from starknet_py.tests.e2e.fixtures.misc import ContractVersion, load_contract


def test_short_strings():
    # docs-shortstring: start
    from starknet_py.cairo.felt import decode_shortstring, encode_shortstring

    # Convert a string literal to its felt value
    encoded = encode_shortstring("myshortstring")
    assert encoded == 0x6D7973686F7274737472696E67

    # Decode a felt into a string
    decoded = decode_shortstring(encoded)
    assert decoded == "myshortstring"
    # docs-shortstring: end


def test_abi_parsing():
    raw_abi_string = json.loads(
        load_contract(contract_name="TestContract", version=ContractVersion.V2)[
            "sierra"
        ]
    )["abi"]
    # docs-serializer: start
    from starknet_py.abi.v2 import AbiParser

    """
    #[external(v0)]
    fn test(ref self: ContractState, ref arg: felt252, arg1: felt252, arg2: felt252) -> felt252 {
        let mut x = self.my_storage_var.read();
        x += 1;
        self.my_storage_var.write(x);
        x + internal_func()
    }
    """

    abi = AbiParser(raw_abi_string).parse()

    from starknet_py.serialization.factory import serializer_for_function_v1

    # You can create serializer for function inputs/outputs by passing Abi.Function object to serializer_for_function
    function_serializer = serializer_for_function_v1(abi.functions["test"])

    # You can call function serializer like you would a normal function
    assert [111, 222, 333] == function_serializer.serialize(arg=111, arg1=222, arg2=333)
    assert [111, 222, 333] == function_serializer.serialize(111, 222, 333)
    assert [111, 222, 333] == function_serializer.serialize(111, 222, arg2=333)

    # You can use deserialized result from function serializer like a tuple
    result = function_serializer.deserialize([1])
    (success,) = result
    assert success == 1
    # docs-serializer: end

    raw_abi_string = json.loads(
        load_contract(contract_name="ERC20", version=ContractVersion.V2)["sierra"]
    )["abi"]
    abi = AbiParser(raw_abi_string).parse()

    # docs-event: start
    from starknet_py.serialization import serializer_for_event

    """
    #[event]
    #[derive(Drop, starknet::Event)]
    struct Approval {
        owner: ContractAddress,
        spender: ContractAddress,
        value: u256,
    }
    """

    # You can create serializer for events by passing Abi.Event object to serializer_for_event
    event_serializer = serializer_for_event(
        abi.events["contracts_v2::erc20::ERC20::Approval"]
    )

    assert [1, 2, 3, 4] == event_serializer.serialize(
        {"owner": 1, "spender": 2, "value": 3 + 4 * 2**128}
    )
    assert {
        "owner": 1,
        "spender": 2,
        "value": 3 + 4 * 2**128,
    } == event_serializer.deserialize([1, 2, 3, 4]).as_dict()
    # docs-event: end



import pytest

from starknet_py.net.models import InvokeV3


@pytest.mark.asyncio
async def test_create_invoke_from_contract(map_contract, account):
    # pylint: disable=import-outside-toplevel
    contract = map_contract

    # docs: start
    from starknet_py.net.client_models import (
        Call,
        ResourceBounds,
        ResourceBoundsMapping,
    )

    # Prepare a call through Contract
    call = contract.functions["put"].prepare_invoke_v3(key=20, value=30)
    assert issubclass(type(call), Call)

    # Crate an Invoke transaction from call
    invoke_transaction = await account.sign_invoke_v3(
        call,
        resource_bounds=ResourceBoundsMapping(
            l1_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
            l1_data_gas=ResourceBounds(
                max_amount=int(1e5), max_price_per_unit=int(1e13)
            ),
            l2_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
        ),
    )
    # docs: end

    assert isinstance(invoke_transaction, InvokeV3)



from sys import platform
from unittest.mock import MagicMock, Mock

import pytest

from starknet_py.constants import STRK_FEE_CONTRACT_ADDRESS
from starknet_py.hash.selector import get_selector_from_name
from starknet_py.net.client_models import Call
from starknet_py.net.full_node_client import FullNodeClient
from starknet_py.net.models import DeclareV3, DeployAccountV3, InvokeV3, StarknetChainId
from starknet_py.net.signer.ledger_signer import BlindSigningModeWarning
from starknet_py.tests.e2e.fixtures.accounts import mint_token_on_devnet
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS_SEPOLIA

LEDGER_ACCOUNT_ADDRESS_SEPOLIA = (
    0x07D2B5E579BB434976E352811D4C3A9DAD7F5966AC2BED4FBBFB7A3B1A0E90DE
)


@pytest.mark.parametrize(
    "transaction",
    [
        InvokeV3(
            version=3,
            signature=[],
            nonce=1,
            resource_bounds=MAX_RESOURCE_BOUNDS_SEPOLIA,
            calldata=[
                1,
                2009894490435840142178314390393166646092438090257831307886760648929397478285,
                232670485425082704932579856502088130646006032362877466777181098476241604910,
                3,
                0x123,
                100,
                0,
            ],
            sender_address=0x123,
            tip=0,
        ),
        DeployAccountV3(
            class_hash=0x123,
            contract_address_salt=0x123,
            constructor_calldata=[1, 2, 3],
            version=3,
            signature=[],
            nonce=0,
            resource_bounds=MAX_RESOURCE_BOUNDS_SEPOLIA,
            tip=0,
        ),
    ],
)
# TODO (#1425): Currently Ledger tests are skipped on Windows due to different Speculos setup.
@pytest.mark.skipif(
    platform == "win32",
    reason="Testing Ledger is skipped on Windows due to different Speculos setup.",
)
def test_clear_sign_transaction(transaction):
    # pylint: disable=redefined-outer-name, unused-import, import-outside-toplevel
    # docs: start
    from starknet_py.net.signer.ledger_signer import LedgerSigner, LedgerSigningMode

    # Create a `LedgerSigner` instance and pass chain id
    signer = LedgerSigner(
        chain_id=StarknetChainId.SEPOLIA,
    )

    # Sign the transaction
    signature = signer.sign_transaction(transaction)
    # docs: end

    assert isinstance(signature, list)
    assert len(signature) == 2
    assert all(isinstance(i, int) for i in signature)
    assert all(i != 0 for i in signature)


@pytest.mark.parametrize(
    "transaction",
    [
        Mock(spec=InvokeV3, calculate_hash=MagicMock(return_value=111)),
        Mock(spec=DeployAccountV3, calculate_hash=MagicMock(return_value=222)),
        Mock(spec=DeclareV3, calculate_hash=MagicMock(return_value=333)),
    ],
)
# TODO (#1425): Currently Ledger tests are skipped on Windows due to different Speculos setup.
@pytest.mark.skipif(
    platform == "win32",
    reason="Testing Ledger is skipped on Windows due to different Speculos setup.",
)
def test_blind_sign_transaction(transaction):
    # pylint: disable=import-outside-toplevel
    from starknet_py.net.signer.ledger_signer import LedgerSigner, LedgerSigningMode

    signer = LedgerSigner(
        chain_id=StarknetChainId.SEPOLIA,
    )
    # docs: start

    # Ledger also allows to blind sign transactions, but keep in mind that blind signing
    # is not recommended. It's unsafe because it lets you approve transactions or
    # messages without seeing their full contents.
    # ⚠️ Use blind signing at your own risk
    signer.signing_mode = LedgerSigningMode.BLIND
    signature = signer.sign_transaction(transaction)
    # docs: end

    assert isinstance(signature, list)
    assert len(signature) == 2
    assert all(isinstance(i, int) for i in signature)
    assert all(i != 0 for i in signature)


# TODO (#1425): Currently Ledger tests are skipped on Windows due to different Speculos setup.
@pytest.mark.skipif(
    platform == "win32",
    reason="Testing Ledger is skipped on Windows due to different Speculos setup.",
)
def test_blind_sign_warning():
    # pylint: disable=import-outside-toplevel, redefined-outer-name
    from starknet_py.net.signer.ledger_signer import LedgerSigner, LedgerSigningMode

    signer = LedgerSigner(
        chain_id=StarknetChainId.SEPOLIA,
    )
    signer.signing_mode = LedgerSigningMode.BLIND

    pattern = (
        "Signing in blind mode is not recommended. It prevents you from verifying "
        "the contents and leaving you vulnerable to unknowingly authorizing malicious transactions. "
        "⚠️ Use at your own risk"
    )
    tx = InvokeV3(
        version=3,
        signature=[],
        nonce=1,
        resource_bounds=MAX_RESOURCE_BOUNDS_SEPOLIA,
        calldata=[
            1,
            2009894490435840142178314390393166646092438090257831307886760648929397478285,
            232670485425082704932579856502088130646006032362877466777181098476241604910,
            3,
            0x123,
            100,
            0,
        ],
        sender_address=0x123,
        tip=0,
    )
    with pytest.warns(BlindSigningModeWarning, match=pattern):
        signer.sign_transaction(tx)


# TODO (#1425): Currently Ledger tests are skipped on Windows due to different Speculos setup.
@pytest.mark.skipif(
    platform == "win32",
    reason="Testing Ledger is skipped on Windows due to different Speculos setup.",
)
def test_create_account_with_ledger_signer():
    # pylint: disable=unused-variable, import-outside-toplevel, redefined-outer-name, reimported
    from starknet_py.net.account.account import Account
    from starknet_py.net.full_node_client import FullNodeClient
    from starknet_py.net.signer.ledger_signer import LedgerSigner

    signer = LedgerSigner(
        chain_id=StarknetChainId.SEPOLIA,
    )

    # docs: start

    client = FullNodeClient(node_url="https://your.node.url")

    # Create an `Account` instance with the ledger signer
    account = Account(
        client=client,
        address=0x1111,
        signer=signer,
        chain=StarknetChainId.SEPOLIA,
    )
    # Now you can use Account as you'd always do
    # docs: end


async def _get_account_balance_strk(client: FullNodeClient, address: int):
    balance = await client.call_contract(
        call=Call(
            to_addr=int(STRK_FEE_CONTRACT_ADDRESS, 16),
            calldata=[address],
            selector=get_selector_from_name("balanceOf"),
        )
    )
    return balance


@pytest.mark.asyncio
# TODO (#1425): Currently Ledger tests are skipped on Windows due to different Speculos setup.
@pytest.mark.skipif(
    platform == "win32",
    reason="Testing Ledger is skipped on Windows due to different Speculos setup.",
)
async def test_deploy_account_and_transfer(client):
    # pylint: disable=import-outside-toplevel, reimported, redefined-outer-name, too-many-locals
    # docs-deploy-account-and-transfer: start
    from starknet_py.contract import Contract
    from starknet_py.hash.address import compute_address
    from starknet_py.net.account.account import Account
    from starknet_py.net.full_node_client import FullNodeClient
    from starknet_py.net.signer.ledger_signer import LedgerSigner

    rpc_client = FullNodeClient(node_url="https://your.node.url")
    # docs-deploy-account-and-transfer: end
    rpc_client = client
    # docs-deploy-account-and-transfer: start
    signer = LedgerSigner(
        chain_id=StarknetChainId.SEPOLIA,
    )
    # argent v0.4.0 class hash
    class_hash = 0x36078334509B514626504EDC9FB252328D1A240E4E948BEF8D0C08DFF45927F
    salt = 1
    calldata = [0, signer.public_key, 1]
    address = compute_address(
        salt=salt,
        class_hash=class_hash,
        constructor_calldata=calldata,
    )
    account = Account(
        address=address,
        client=rpc_client,
        signer=signer,
        chain=StarknetChainId.SEPOLIA,
    )

    # Remember to prefund the account
    # docs-deploy-account-and-transfer: end
    # Here we prefund the devnet account for test purposes
    await mint_token_on_devnet(
        url=rpc_client.url.replace("/rpc", ""),
        address=address,
        amount=5000000000000000000000,
        unit="FRI",
    )
    # docs-deploy-account-and-transfer: start
    signed_tx = await account.sign_deploy_account_v3(
        class_hash=class_hash,
        contract_address_salt=salt,
        constructor_calldata=calldata,
        auto_estimate=True,
    )

    await rpc_client.deploy_account(signed_tx)

    recipient_address = 0x123
    # docs-deploy-account-and-transfer: end
    recipient_balance_before = (
        await _get_account_balance_strk(rpc_client, recipient_address)
    )[0]
    # docs-deploy-account-and-transfer: start
    contract = await Contract.from_address(
        provider=account, address=STRK_FEE_CONTRACT_ADDRESS
    )
    invocation = await contract.functions["transfer"].invoke_v3(
        recipient_address, 100, auto_estimate=True
    )
    await invocation.wait_for_acceptance()
    # docs-deploy-account-and-transfer: end
    recipient_balance_after = (
        await _get_account_balance_strk(rpc_client, recipient_address)
    )[0]

    assert recipient_balance_before + 100 == recipient_balance_after


@pytest.mark.skip(reason="TODO(#1637)")
@pytest.mark.asyncio
# TODO (#1425): Currently Ledger tests are skipped on Windows due to different Speculos setup.
@pytest.mark.skipif(
    platform == "win32",
    reason="Testing Ledger is skipped on Windows due to different Speculos setup.",
)
async def test_invoke_v3_long_calldata(client_fork_mode):
    # pylint: disable=import-outside-toplevel, redefined-outer-name, reimported
    from starknet_py.contract import Contract
    from starknet_py.net.account.account import Account
    from starknet_py.net.signer.ledger_signer import LedgerSigner

    # Contract deployed on Sepolia
    contract_address = (
        0x042C25F2DD9C4AA010E7A4ADA1BFB1C99E5DBEA2850C59D8FD9F59F554CC268E
    )

    signer = LedgerSigner(
        chain_id=StarknetChainId.SEPOLIA,
    )

    account = Account(
        address=LEDGER_ACCOUNT_ADDRESS_SEPOLIA,
        client=client_fork_mode,
        signer=signer,
        chain=StarknetChainId.SEPOLIA,
    )

    contract = await Contract.from_address(provider=account, address=contract_address)

    # `fn_with_many_args` accepts 17 arguments
    args = list(range(1, 18))

    invocation = await contract.functions["fn_with_many_args"].invoke_v3(
        *args,
        auto_estimate=True,
    )
    await invocation.wait_for_acceptance()


@pytest.mark.skip(reason="TODO(#1637)")
@pytest.mark.asyncio
# TODO (#1425): Currently Ledger tests are skipped on Windows due to different Speculos setup.
@pytest.mark.skipif(
    platform == "win32",
    reason="Testing Ledger is skipped on Windows due to different Speculos setup.",
)
async def test_deploy_account_v3_long_calldata(client_fork_mode):
    # pylint: disable=import-outside-toplevel, redefined-outer-name, reimported
    from starknet_py.contract import Contract
    from starknet_py.net.account.account import Account
    from starknet_py.net.signer.ledger_signer import LedgerSigner

    # Contract declared on Sepolia
    class_hash = 0x040ACE4954F5F7D8BF202A87EAD2AD4BA77F245740A35DD11AFD6912DEB08ABF

    signer = LedgerSigner(
        chain_id=StarknetChainId.SEPOLIA,
    )

    account = Account(
        address=LEDGER_ACCOUNT_ADDRESS_SEPOLIA,
        client=client_fork_mode,
        signer=signer,
        chain=StarknetChainId.SEPOLIA,
    )

    # constructor accepts 17 arguments
    constructor_args = list(range(1, 18))

    deploy_result = await Contract.deploy_contract_v3(
        account=account,
        class_hash=class_hash,
        constructor_args=constructor_args,
        auto_estimate=True,
    )
    await deploy_result.wait_for_acceptance()



import pytest

from starknet_py.serialization.data_serializers.byte_array_serializer import (
    ByteArraySerializer,
)

byte_array_serializer = ByteArraySerializer()


@pytest.mark.parametrize(
    "value, serialized_value",
    [
        ("", [0, 0, 0]),
        ("hello", [0, 0x68656C6C6F, 5]),
        (
            "Long string, more than 31 characters.",
            [
                1,
                0x4C6F6E6720737472696E672C206D6F7265207468616E203331206368617261,
                0x63746572732E,
                6,
            ],
        ),
    ],
)
def test_values(value, serialized_value):
    serialized = byte_array_serializer.serialize(value)
    deserialized = byte_array_serializer.deserialize(serialized_value)

    assert deserialized == value
    assert serialized == serialized_value



import pytest

from starknet_py.serialization._calldata_reader import CalldataReader, OutOfBoundsError


def test_consuming_no_calldata():
    reader = CalldataReader(list(range(0, 1)))

    with pytest.raises(ValueError, match="size must be greater than 0"):
        reader.read(0)


def test_consuming_calldata():
    reader = CalldataReader(list(range(0, 100)))

    assert reader.remaining_len == 100

    assert reader.read(10) == list(range(0, 10))
    assert reader.remaining_len == 90

    assert reader.read(5) == list(range(10, 15))
    assert reader.remaining_len == 85

    assert reader.read(35) == list(range(15, 50))
    assert reader.remaining_len == 50

    with pytest.raises(
        OutOfBoundsError, match="Requested 51 elements, 50 available."
    ) as err_info:
        reader.read(51)

    assert err_info.value.position == 50
    assert err_info.value.remaining_len == 50
    assert err_info.value.requested_size == 51

    # Nothing was consumed when requested too much above
    assert reader.read(50) == list(range(50, 100))
    assert reader.remaining_len == 0


def test_empty_calldata():
    reader = CalldataReader([])

    assert reader.remaining_len == 0

    with pytest.raises(
        OutOfBoundsError, match="Requested 10 elements, 0 available."
    ) as err_info:
        reader.read(10)

    assert err_info.value.position == 0
    assert err_info.value.remaining_len == 0
    assert err_info.value.requested_size == 10
